# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Tests for TestTcodDbExporter
"""
import unittest

from aiida.backends.testbase import AiidaTestCase
from aiida.common.links import LinkType

def has_nwchem_plugin():
    from aiida.common.exceptions import MissingEntryPointError
    from aiida.plugins.entry_point import load_entry_point
    from aiida.tools.dbexporters.tcod_plugins import BaseTcodtranslator

    try:
        load_entry_point('aiida.tools.dbexporters.tcod_plugins', 'nwchem.nwcpymatgen')
    except MissingEntryPointError:
        return False

    return True


class FakeObject(object):
    """
    A wrapper for dictionary, which can be used instead of object.
    Example use case: fake Calculation object ``calc``, having keys
    ``inp`` and ``out`` to access also fake NodeInputManager and
    NodeOutputManager.
    """

    def __init__(self, dictionary):
        self._dictionary = dictionary

    def __getattr__(self, name):
        if isinstance(self._dictionary[name], dict):
            return FakeObject(self._dictionary[name])
        else:
            return self._dictionary[name]


class TestTcodDbExporter(AiidaTestCase):
    """
    Tests for TcodDbExporter class.
    """
    from aiida.orm.data.structure import has_ase, has_spglib
    from aiida.orm.data.cif import has_pycifrw

    def test_contents_encoding_1(self):
        """
        Testing the logic of choosing the encoding and the process of
        encoding contents.
        """
        from aiida.tools.dbexporters.tcod import cif_encode_contents
        self.assertEquals(cif_encode_contents('simple line')[1],
                          None)
        self.assertEquals(cif_encode_contents(' ;\n ;')[1],
                          None)
        self.assertEquals(cif_encode_contents(';\n'),
                          ('=3B\n', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('line\n;line'),
                          ('line\n=3Bline', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('tabbed\ttext'),
                          ('tabbed=09text', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('angstrom Å'),
                          ('angstrom =C3=85', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('.'),
                          ('=2E', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('?'),
                          ('=3F', 'quoted-printable'))
        self.assertEquals(cif_encode_contents('.?'), ('.?', None))
        # This one is particularly tricky: a long line is folded by the QP
        # and the semicolon sign becomes the first character on a new line.
        self.assertEquals(cif_encode_contents(
            "Å{};a".format("".join("a" for i in range(0, 69)))),
            ('=C3=85aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
             'aaaaaaaaaaaaaaaaaaaaaaaaaaaaa=\n=3Ba',
             'quoted-printable'))
        self.assertEquals(cif_encode_contents('angstrom ÅÅÅ'),
                          ('YW5nc3Ryb20gw4XDhcOF', 'base64'))
        self.assertEquals(cif_encode_contents(
            "".join("a" for i in range(0, 2048)))[1],
                          None)
        self.assertEquals(cif_encode_contents(
            "".join("a" for i in range(0, 2049)))[1],
                          'quoted-printable')
        self.assertEquals(cif_encode_contents('datatest')[1], None)
        self.assertEquals(cif_encode_contents('data_test')[1], 'base64')

    def test_collect_files(self):
        """
        Testing the collection of files from file tree.
        """
        from aiida.tools.dbexporters.tcod import _collect_files
        from aiida.common.folders import SandboxFolder
        import StringIO

        sf = SandboxFolder()
        sf.get_subfolder('out', create=True)
        sf.get_subfolder('pseudo', create=True)
        sf.get_subfolder('save', create=True)
        sf.get_subfolder('save/1', create=True)
        sf.get_subfolder('save/2', create=True)

        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'aiida.in')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'aiida.out')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, '_aiidasubmit.sh')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, '_.out')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'out/out')
        f = StringIO.StringIO("test")
        sf.create_file_from_filelike(f, 'save/1/log.log')

        md5 = '098f6bcd4621d373cade4e832627b4f6'
        sha1 = 'a94a8fe5ccb19ba61c4c0873d391e987982fbbd3'
        self.assertEquals(
            _collect_files(sf.abspath),
            [{'name': '_.out', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': '_aiidasubmit.sh', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'aiida.in', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'aiida.out', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'out/', 'type': 'folder'},
             {'name': 'out/out', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'pseudo/', 'type': 'folder'},
             {'name': 'save/', 'type': 'folder'},
             {'name': 'save/1/', 'type': 'folder'},
             {'name': 'save/1/log.log', 'contents': 'test', 'md5': md5,
              'sha1': sha1, 'type': 'file'},
             {'name': 'save/2/', 'type': 'folder'}])

    @unittest.skipIf(not has_ase(), "Unable to import ase")
    @unittest.skipIf(not has_spglib(), "Unable to import spglib")
    @unittest.skipIf(not has_pycifrw(), "Unable to import PyCifRW")
    @unittest.skipIf(not has_nwchem_plugin(), "NWChem plugin is not installed")
    def test_cif_structure_roundtrip(self):
        from aiida.tools.dbexporters.tcod import export_cif, export_values
        from aiida.orm import Code
        from aiida.orm import JobCalculation
        from aiida.orm.data.cif import CifData
        from aiida.orm.data.parameter import ParameterData
        from aiida.orm.data.upf import UpfData
        from aiida.orm.data.folder import FolderData
        from aiida.common.folders import SandboxFolder
        from aiida.common.datastructures import calc_states
        import tempfile

        with tempfile.NamedTemporaryFile() as f:
            f.write('''
                data_test
                _cell_length_a    10
                _cell_length_b    10
                _cell_length_c    10
                _cell_angle_alpha 90
                _cell_angle_beta  90
                _cell_angle_gamma 90
                loop_
                _atom_site_label
                _atom_site_fract_x
                _atom_site_fract_y
                _atom_site_fract_z
                C 0 0 0
                O 0.5 0.5 0.5
            ''')
            f.flush()
            a = CifData(file=f.name)

        c = a._get_aiida_structure()
        c.store()
        pd = ParameterData()

        code = Code(local_executable='test.sh')
        with tempfile.NamedTemporaryFile() as f:
            f.write("#/bin/bash\n\necho test run\n")
            f.flush()
            code.add_path(f.name, 'test.sh')

        code.store()

        calc = JobCalculation(computer=self.computer)
        calc.set_resources({'num_machines': 1,
                            'num_mpiprocs_per_machine': 1})
        calc.add_link_from(code, "code")
        calc.set_environment_variables({'PATH': '/dev/null', 'USER': 'unknown'})

        with tempfile.NamedTemporaryFile(prefix="Fe") as f:
            f.write("<UPF version=\"2.0.1\">\nelement=\"Fe\"\n")
            f.flush()
            upf = UpfData(file=f.name)
            upf.store()
            calc.add_link_from(upf, "upf")

        with tempfile.NamedTemporaryFile() as f:
            f.write("data_test")
            f.flush()
            cif = CifData(file=f.name)
            cif.store()
            calc.add_link_from(cif, "cif")

        calc.store()
        calc._set_state(calc_states.SUBMITTING)
        with SandboxFolder() as f:
            calc._store_raw_input_folder(f.abspath)

        fd = FolderData()
        with open(fd._get_folder_pathsubfolder.get_abs_path(
                calc._SCHED_OUTPUT_FILE), 'w') as f:
            f.write("standard output")
            f.flush()

        with open(fd._get_folder_pathsubfolder.get_abs_path(
                calc._SCHED_ERROR_FILE), 'w') as f:
            f.write("standard error")
            f.flush()

        fd.store()
        fd.add_link_from(calc, calc._get_linkname_retrieved(), LinkType.CREATE)

        pd.add_link_from(calc, "calc", LinkType.CREATE)
        pd.store()

        with self.assertRaises(ValueError):
            export_cif(c, parameters=pd)

        c.add_link_from(calc, "calc", LinkType.CREATE)
        export_cif(c, parameters=pd)

        values = export_values(c, parameters=pd)
        values = values['0']

        self.assertEquals(values['_tcod_computation_environment'],
                          ['PATH=/dev/null\nUSER=unknown'])
        self.assertEquals(values['_tcod_computation_command'],
                          ['cd 1; ./_aiidasubmit.sh'])


    @unittest.skipIf(not has_nwchem_plugin(), "NWChem plugin is not installed")
    def test_nwcpymatgen_translation(self):
        from tcodexporter import FakeObject
        from aiida.orm.data.parameter import ParameterData
        from aiida.plugins.entry_point import load_entry_point
        from aiida.tools.dbexporters.tcod import translate_calculation_specific_values

        nwchem_plugin = load_entry_point('aiida.tools.dbexporters.tcod_plugins', 'nwchem.nwcpymatgen')

        calc = FakeObject({
            "out": {"output":
                ParameterData(dict={
                    "basis_set": {
                        "H": {
                            "description": "6-31g",
                            "functions": "2",
                            "shells": "2",
                            "types": "2s"
                        },
                        "O": {
                            "description": "6-31g",
                            "functions": "9",
                            "shells": "5",
                            "types": "3s2p"
                        }
                    },
                    "corrections": {},
                    "energies": [
                        -2057.99011937535
                    ],
                    "errors": [],
                    "frequencies": None,
                    "has_error": False,
                    "job_type": "NWChem SCF Module"
                }),
                "job_info": ParameterData(dict={
                    "0 permanent": ".",
                    "0 scratch": ".",
                    "argument  1": "aiida.in",
                    "compiled": "Sun_Dec_22_04:02:59_2013",
                    "data base": "./aiida.db",
                    "date": "Mon May 11 17:10:07 2015",
                    "ga revision": "10379",
                    "global": "200.0 Mbytes (distinct from heap & stack)",
                    "hardfail": "no",
                    "heap": "100.0 Mbytes",
                    "hostname": "theospc11",
                    "input": "aiida.in",
                    "nproc": "6",
                    "nwchem branch": "6.3",
                    "nwchem revision": "24277",
                    "prefix": "aiida.",
                    "program": "/usr/bin/nwchem",
                    "source": "/build/buildd/nwchem-6.3+r1",
                    "stack": "100.0 Mbytes",
                    "status": "startup",
                    "time left": "-1s",
                    "total": "400.0 Mbytes",
                    "verify": "yes",
                })
            }})
        res = translate_calculation_specific_values(calc, nwchem_plugin)
        self.assertEquals(res, {
            '_tcod_software_package': 'NWChem',
            '_tcod_software_package_version': '6.3',
            '_tcod_software_package_compilation_date': '2013-12-22T04:02:59',
            '_atom_type_symbol': ['H', 'O'],
            '_dft_atom_basisset': ['6-31g', '6-31g'],
            '_dft_atom_type_valence_configuration': ['2s', '3s2p'],
        })

    @unittest.skipIf(not has_ase(), "Unable to import ase")
    @unittest.skipIf(not has_spglib(), "Unable to import spglib")
    @unittest.skipIf(not has_pycifrw(), "Unable to import PyCifRW")
    @unittest.skipIf(not has_nwchem_plugin(), "NWChem plugin is not installed")
    def test_inline_export(self):
        from aiida.orm.data.cif import CifData
        from aiida.tools.dbexporters.tcod import export_values
        import tempfile

        with tempfile.NamedTemporaryFile() as f:
            f.write('''
                data_test
                _cell_length_a    10
                _cell_length_b    10
                _cell_length_c    10
                _cell_angle_alpha 90
                _cell_angle_beta  90
                _cell_angle_gamma 90
                loop_
                _atom_site_label
                _atom_site_fract_x
                _atom_site_fract_y
                _atom_site_fract_z
                C 0 0 0
                O 0.5 0.5 0.5
            ''')
            f.flush()
            a = CifData(file=f.name)

        s = a._get_aiida_structure(store=True)
        val = export_values(s)
        script = val.first_block()['_tcod_file_contents'][1]
        function = '_get_aiida_structure_ase_inline'
        self.assertNotEqual(script.find(function), script.rfind(function))

    @unittest.skipIf(not has_ase(), "Unable to import ase")
    @unittest.skipIf(not has_spglib(), "Unable to import spglib")
    @unittest.skipIf(not has_pycifrw(), "Unable to import PyCifRW")
    def test_symmetry_reduction(self):
        from aiida.orm.data.structure import StructureData
        from aiida.tools.dbexporters.tcod import export_values
        from ase import Atoms

        a = Atoms('BaTiO3', cell=(4., 4., 4.))
        a.set_scaled_positions(
            ((0.0, 0.0, 0.0),
             (0.5, 0.5, 0.5),
             (0.5, 0.5, 0.0),
             (0.5, 0.0, 0.5),
             (0.0, 0.5, 0.5),
             )
        )

        a.set_chemical_symbols(['Ba', 'Ti', 'O', 'O', 'O'])
        val = export_values(StructureData(ase=a), reduce_symmetry=True, store=True)['0']
        self.assertEqual(val['_atom_site_label'], ['Ba1', 'Ti1', 'O1'])
        self.assertEqual(val['_symmetry_space_group_name_H-M'], 'Pm-3m')
        self.assertEqual(val['_symmetry_space_group_name_Hall'], '-P 4 2 3')

    def test_cmdline_parameters(self):
        """
        Ensuring that neither extend_with_cmdline_parameters() nor
        deposition_cmdline_parameters() set default parameters.
        """
        from aiida.tools.dbexporters.tcod \
            import extend_with_cmdline_parameters, \
            deposition_cmdline_parameters
        import argparse

        parser = argparse.ArgumentParser()
        extend_with_cmdline_parameters(parser)
        options = vars(parser.parse_args(args=[]))

        for key in options.keys():
            if options[key] is None:
                options.pop(key)

        self.assertEqual(options, {})

        parser = argparse.ArgumentParser()
        deposition_cmdline_parameters(parser)
        options = vars(parser.parse_args(args=[]))

        for key in options.keys():
            if options[key] is None:
                options.pop(key)

        self.assertEqual(options, {})

    @unittest.skipIf(not has_ase(), "Unable to import ase")
    @unittest.skipIf(not has_spglib(), "Unable to import spglib")
    @unittest.skipIf(not has_pycifrw(), "Unable to import PyCifRW")
    def test_export_trajectory(self):
        from aiida.orm.data.structure import StructureData
        from aiida.orm.data.array.trajectory import TrajectoryData
        from aiida.tools.dbexporters.tcod import export_values

        cells = [
            [[2., 0., 0., ],
             [0., 2., 0., ],
             [0., 0., 2., ]],
            [[3., 0., 0., ],
             [0., 3., 0., ],
             [0., 0., 3., ]]
        ]
        symbols = [['H', 'O', 'C'], ['H', 'O', 'C']]
        positions = [
            [[0., 0., 0.],
             [0.5, 0.5, 0.5],
             [1.5, 1.5, 1.5]],
            [[0., 0., 0.],
             [0.75, 0.75, 0.75],
             [1.25, 1.25, 1.25]]
        ]
        structurelist = []
        for i in range(0, 2):
            struct = StructureData(cell=cells[i])
            for j, symbol in enumerate(symbols[i]):
                struct.append_atom(symbols=symbol, position=positions[i][j])
            structurelist.append(struct)

        td = TrajectoryData(structurelist=structurelist)

        with self.assertRaises(ValueError):
            # Trajectory index is not specified
            v = export_values(td)

        expected_tags = [
            '_atom_site_fract_x',
            '_atom_site_fract_y',
            '_atom_site_fract_z',
            '_atom_site_label',
            '_atom_site_type_symbol',
            '_audit_conform_dict_location',
            '_audit_conform_dict_name',
            '_audit_conform_dict_version',
            '_audit_creation_method',
            '_cell_angle_alpha',
            '_cell_angle_beta',
            '_cell_angle_gamma',
            '_cell_length_a',
            '_cell_length_b',
            '_cell_length_c',
            '_chemical_formula_sum',
            '_symmetry_equiv_pos_as_xyz',
            '_symmetry_int_tables_number',
            '_symmetry_space_group_name_h-m',
            '_symmetry_space_group_name_hall'
        ]

        tcod_file_tags = [
            '_tcod_content_encoding_id',
            '_tcod_content_encoding_layer_id',
            '_tcod_content_encoding_layer_type',
            '_tcod_file_content_encoding',
            '_tcod_file_contents',
            '_tcod_file_id',
            '_tcod_file_md5sum',
            '_tcod_file_name',
            '_tcod_file_role',
            '_tcod_file_sha1sum',
            '_tcod_file_uri',
        ]

        # Not stored and not to be stored:
        v = export_values(td, trajectory_index=1)
        self.assertEqual(sorted(v['0'].keys()), expected_tags)

        # Stored, but not expected to be stored:
        td = TrajectoryData(structurelist=structurelist)
        td.store()
        v = export_values(td, trajectory_index=1)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags + tcod_file_tags)

        # Not stored, but expected to be stored:
        td = TrajectoryData(structurelist=structurelist)
        v = export_values(td, trajectory_index=1, store=True)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags + tcod_file_tags)
        
        # Both stored and expected to be stored:
        td = TrajectoryData(structurelist=structurelist)
        td.store()
        v = export_values(td, trajectory_index=1, store=True)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags + tcod_file_tags)

        # Stored, but asked not to include DB dump:
        td = TrajectoryData(structurelist=structurelist)
        td.store()
        v = export_values(td, trajectory_index=1,
                          dump_aiida_database=False)
        self.assertEqual(sorted(v['0'].keys()),
                         expected_tags)

    def test_contents_encoding_2(self):
        """
        Testing the logic of choosing the encoding and the process of
        encoding contents.
        """
        from aiida.tools.dbexporters.tcod import decode_textfield

        def check_ncr(self, inp, out):
            from aiida.tools.dbexporters.tcod import (encode_textfield_ncr,
                                                      decode_textfield_ncr)
            encoded = encode_textfield_ncr(inp)
            decoded = decode_textfield_ncr(out)
            decoded_universal = decode_textfield(out, 'ncr')
            self.assertEquals(encoded, out)
            self.assertEquals(decoded, inp)
            self.assertEquals(decoded_universal, inp)

        def check_quoted_printable(self, inp, out):
            from aiida.tools.dbexporters.tcod import (encode_textfield_quoted_printable,
                                                      decode_textfield_quoted_printable)
            encoded = encode_textfield_quoted_printable(inp)
            decoded = decode_textfield_quoted_printable(out)
            decoded_universal = decode_textfield(out, 'quoted-printable')
            self.assertEquals(encoded, out)
            self.assertEquals(decoded, inp)
            self.assertEquals(decoded_universal, inp)

        def check_base64(self, inp, out):
            from aiida.tools.dbexporters.tcod import (encode_textfield_base64,
                                                      decode_textfield_base64)
            encoded = encode_textfield_base64(inp)
            decoded = decode_textfield_base64(out)
            decoded_universal = decode_textfield(out, 'base64')
            self.assertEquals(encoded, out)
            self.assertEquals(decoded, inp)
            self.assertEquals(decoded_universal, inp)

        def check_gzip_base64(self, text):
            from aiida.tools.dbexporters.tcod import (encode_textfield_gzip_base64,
                                                      decode_textfield_gzip_base64)
            encoded = encode_textfield_gzip_base64(text)
            decoded = decode_textfield_gzip_base64(encoded)
            decoded_universal = decode_textfield(encoded, 'gzip+base64')
            self.assertEquals(text, decoded)
            self.assertEquals(text, decoded_universal)

        check_ncr(self, '.', '&#46;')
        check_ncr(self, '?', '&#63;')
        check_ncr(self, ';\n', '&#59;\n')
        check_ncr(self, 'line\n;line', 'line\n&#59;line')
        check_ncr(self, 'tabbed\ttext', 'tabbed&#9;text')
        check_ncr(self, 'angstrom Å', 'angstrom &#195;&#133;')
        check_ncr(self, '<html>&#195;&#133;</html>',
                 '<html>&#38;#195;&#38;#133;</html>')

        check_quoted_printable(self, '.', '=2E')
        check_quoted_printable(self, '?', '=3F')
        check_quoted_printable(self, ';\n', '=3B\n')
        check_quoted_printable(self, 'line\n;line', 'line\n=3Bline')
        check_quoted_printable(self, 'tabbed\ttext', 'tabbed=09text')
        check_quoted_printable(self, 'angstrom Å', 'angstrom =C3=85')
        check_quoted_printable(self, 'line\rline\x00', 'line=0Dline=00')
        # This one is particularly tricky: a long line is folded by the QP
        # and the semicolon sign becomes the first character on a new line.
        check_quoted_printable(self,
                              "Å{};a".format("".join("a" for i in range(0, 69))),
                              '=C3=85aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
                              'aaaaaaaaaaaaaaaaaaaaaaaaaaaaa=\n=3Ba')

        check_base64(self, 'angstrom ÅÅÅ', 'YW5nc3Ryb20gw4XDhcOF')
        check_gzip_base64(self, 'angstrom ÅÅÅ')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import sys
import os

import aiida
from aiida.backends.utils import load_dbenv, is_dbenv_loaded
from aiida.cmdline import pass_to_django_manage, execname
from aiida.cmdline.baseclass import VerdiCommandWithSubcommands
from aiida.cmdline.utils import decorators
from aiida.common.exceptions import InternalError, TestsNotAllowedError



def applyfunct_len(value):
    """
    Return the length of an object.
    """
    try:
        return len(value)
    except Exception as e:
        raise InternalError(e, 'Error in function len, probably the object has no "len"?')


def applyfunct_keys(value):
    """
    Return the keys of a dictionary.
    """
    try:
        return value.keys()
    except Exception as e:
        raise InternalError(e, 'Error in function keys, probably not a dict?')


def apply_function(function, value):
    """
    The function must be defined in this file and be in the format
    applyfunct_FUNCNAME
    where FUNCNAME is the string passed as the parameter 'function';
    applyfunct_FUNCNAME will accept only one parameter ('value') and
    return an appropriate value.
    """
    function_prefix = "applyfunct_"
    if function is None:
        return value
    else:
        try:
            return globals()[function_prefix + function](value)
        except KeyError as e:
            # Raising an InternalError means that a default value is printed
            # if no function exists.
            # Instead, raising a ValueError will always get printed even if
            # a default value is printed, that is what one wants
            # raise InternalError(
            #                real_exception=e, message =
            raise ValueError(
                "o such function %s. Available functions are: %s." %
                (function,
                 ", ".join(i[len(function_prefix):] for i in globals() if i.startswith(function_prefix))))


class Devel(VerdiCommandWithSubcommands):
    """
    AiiDA commands for developers

    Provides a set of tools for developers. For instance, it allows to run
    the django tests for the db application and the unittests of
    the AiiDA modules.

    If you want to limit the tests to a specific subset of modules,
    pass them as parameters.

    An invalid parameter will make the code print the list of all
    valid parameters.

    Note: the test called 'db' will run all db.* tests.
    """
    base_allowed_test_folders = [
        'aiida.scheduler',
        'aiida.transport',
        'aiida.common',
        'aiida.utils',
        'aiida.control',
        'aiida.cmdline.tests'
    ]

    _dbrawprefix = "db"
    _dbprefix = _dbrawprefix + "."

    def __init__(self, *args, **kwargs):
        from aiida.backends.tests import get_db_test_names
        from aiida.backends import settings

        load_dbenv()
        db_test_list = get_db_test_names()

        super(Devel, self).__init__(*args, **kwargs)

        self.valid_subcommands = {
            'tests': (self.run_tests, self.complete_tests),
            'query': (self.run_query, self.complete_none),  # For the moment, no completion
            'setproperty': (self.run_setproperty, self.complete_properties),
            'getproperty': (self.run_getproperty, self.complete_properties),
            'delproperty': (self.run_delproperty, self.complete_properties),
            'describeproperties': (self.run_describeproperties, self.complete_none),
            'listproperties': (self.run_listproperties, self.complete_none),
            'play': (self.run_play, self.complete_none),
            'getresults': (self.calculation_getresults, self.complete_none),
            'run_daemon': (self.run_daemon, self.complete_none)
        }

        # The content of the dict is:
        # None for a simple folder test
        #   a list of strings for db tests, one for each test to run
        self.allowed_test_folders = {}
        for k in self.base_allowed_test_folders:
            self.allowed_test_folders[k] = None

        for dbtest in db_test_list:
            self.allowed_test_folders["{}{}".format(self._dbprefix, dbtest)] = [dbtest]
        self.allowed_test_folders[self._dbrawprefix] = db_test_list

    def complete_properties(self, subargs_idx, subargs):
        """
        I complete with subargs that were not used yet.
        """
        from aiida.common.setup import _property_table

        if subargs_idx == 0:
            return " ".join(_property_table.keys())
        else:
            return ""

    def run_describeproperties(self, *args):
        """
        List all valid properties that can be stored in the AiiDA config file.

        Only properties listed in the ``_property_table`` of
        ``aida.common.setup`` can be used.
        """
        from aiida.common.setup import _property_table, _NoDefaultValue

        if args:
            print >> sys.stderr, ("No parameters allowed for {}".format(
                self.get_full_command_name()))
            sys.exit(1)

        for prop in sorted(_property_table.keys()):
            if _property_table[prop][4] is None:
                valid_vals_str = ""
            else:
                valid_vals_str = " Valid values: {}.".format(",".join(
                    str(_) for _ in _property_table[prop][4]))
            if isinstance(_property_table[prop][3], _NoDefaultValue):
                def_val_string = ""
            else:
                def_val_string = " (default: {})".format(
                    _property_table[prop][3])
            print "* {} ({}): {}{}{}".format(prop, _property_table[prop][1],
                                             _property_table[prop][2],
                                             def_val_string, valid_vals_str)

    def calculation_getresults(self, *args):
        """
        Routine to get a list of results of a set of calculations, still
        under development.
        """
        from aiida.common.exceptions import AiidaException
        load_dbenv()
        from aiida.orm import JobCalculation as OrmCalculation
        from aiida.orm.utils import load_node


        class InternalError(AiidaException):
            def __init__(self, real_exception, message):
                self.real_exception = real_exception
                self.message = message


        def get_suggestions(key, correct_keys):
            import difflib
            import string

            similar_kws = difflib.get_close_matches(key, correct_keys)
            if len(similar_kws) == 1:
                return "(Maybe you wanted to specify %s?)" % similar_kws[0]
            elif len(similar_kws) > 1:
                return "(Maybe you wanted to specify one of these: %s?)" % string.join(similar_kws, ', ')
            else:
                return "(No similar keywords found...)"

        # define a function to retrieve the data from the dictionary
        def key_finder(in_dict, the_keys):
            parent_dict = in_dict
            parent_dict_name = '<root level>'
            for new_key in the_keys:
                try:
                    parent_dict = parent_dict[new_key]
                    parent_dict_name = new_key
                except KeyError as e:
                    raise InternalError(e, "Unable to find the key '%s' in '%s' %s" % (
                        new_key, parent_dict_name, get_suggestions(new_key, parent_dict.keys())))
                except Exception as e:
                    if e.__class__ is not InternalError:
                        raise InternalError(e,
                                            "Error retrieving the key '%s' withing '%s', maybe '%s' is not a dict?" % (
                                                the_keys[1], the_keys[0], the_keys[0]))
                    else:
                        raise
            return parent_dict


        def index_finder(data, indices, list_name):
            if not indices:
                return data

            parent_data = data
            parent_name = list_name
            for idx in indices:
                try:
                    index = int(idx)
                    parent_data = parent_data[index]
                    parent_name += ":%s" % index
                except ValueError as e:
                    raise InternalError(e, "%s is not a valid integer (in %s)." % (idx, parent_name))
                except IndexError as e:
                    raise InternalError(e, "Index %s is out of bounds, length of list %s is %s" % (
                        index, parent_name, len(parent_data)))
                except Exception as e:
                    raise InternalError(e, "Invalid index! Maybe %s is not a list?" % parent_name)
            return parent_data


        if not args:
            print >> sys.stderr, "Pass some parameters."
            sys.exit(1)

        # I convert it to a list
        arguments = list(args)

        try:
            sep_idx = arguments.index('--')
        except ValueError:
            print >> sys.stderr, "Separate parameter keys from job-ids with a --!"
            sys.exit(1)

        keys_to_retrieve = arguments[:sep_idx]
        job_list_str = arguments[sep_idx + 1:]

        try:
            job_list = [int(i) for i in job_list_str]
        except ValueError:
            print >> sys.stderr, "All PKs after -- must be valid integers."
            sys.exit(1)


        sep = '\t'  # Default separator: a tab character

        try:
            idx = keys_to_retrieve.index('-s')

            # First pop removes the -s parameter, second pop (now with the same
            # index) retrieves the asked separator. Raises IndexError if nothing
            # is provided after -s.
            _ = keys_to_retrieve.pop(idx)
            sep = keys_to_retrieve.pop(idx)
        except IndexError:
            print >> sys.stderr, "After -s you have to pass a valid separator!"
            sys.exit(1)
        except ValueError:
            # No -s found, use default separator
            pass

        print_header = True  # Default: print the header
        try:
            keys_to_retrieve.remove('--no-header')
            # If here, the key was found: I remove it from the list and set the
            # the print_header flag
            print_header = False
        except ValueError:
            # No --no-header found: pass
            pass

        # check if there is at least one thing to do
        if not job_list:
            print >> sys.stderr, "Failed recognizing a calculation PK."
            sys.exit(1)
        if not keys_to_retrieve:
            print  >> sys.stderr, "Failed recognizing a key to parse."
            sys.exit(1)

        # load the data
        if print_header:
            print "## Job list: %s" % " ".join(job_list_str)
            print "#" + sep.join(str(i) for i in keys_to_retrieve)
        for job in job_list:
            values_to_print = []
            in_found = True
            out_found = True
            c = load_node(job, parent_class=OrmCalculation)
            try:
                i = c.inp.parameters.get_dict()
            except AttributeError:
                i = {}
                in_found = False
            try:
                o = c.out.output_parameters.get_dict()
            except AttributeError:
                out_found = False
                o = {}

            io = {'extras': c.get_extras(), 'attrs': c.get_attrs(),
                  'i': i, 'o': o, 'pk': job, 'label': c.label,
                  'desc': c.description, 'state': c.get_state(),
                  'sched_state': c.get_scheduler_state(),
                  'owner': c.dbnode.user.email}

            try:
                for datakey_full in keys_to_retrieve:
                    split_for_def = datakey_full.split('=')
                    datakey_raw = split_for_def[0]  # Still can contain the function to apply
                    def_val = "=".join(split_for_def[1:]) if len(split_for_def) > 1 else None

                    split_for_func = datakey_raw.split('/')
                    datakey = split_for_func[0]
                    function_to_apply = "/".join(split_for_func[1:]) if len(split_for_func) > 1 else None

                    key_values = [str(i) for i in datakey.split(':')[0].split('.')]
                    indices = [int(i) for i in datakey.split(':')[1:]]  # empty list for simple variables.
                    try:
                        value_key = key_finder(io, key_values)
                        value = index_finder(value_key, indices, list_name=datakey.split(':')[0])
                        # Will do the correct thing if no function is supplied
                        # and function is therefore None
                        value = apply_function(function=function_to_apply, value=value)

                    except InternalError as e:
                        # For any internal error, set the value to the default if
                        # provided, otherwise re-raise the exception
                        if def_val is not None:
                            value = def_val
                        else:
                            if not out_found:
                                if in_found:
                                    e.message += " [NOTE: No output JSON could be found]"
                                else:
                                    e.message += " [NOTE: No output JSON nor input JSON could be found]"
                            raise e

                    values_to_print.append(value)

                    # Not needed: now we can ask explicitly for a job number using the flag
                    # 'jobnum'
                    # values_to_print.append("# %s" % str(job))
                # print on screen
                print sep.join(str(i) for i in values_to_print)
            except InternalError as e:
                print >> sys.stderr, e.message
            except Exception as e:
                print >> sys.stderr, "# Error loading job # %s (%s): %s" % (job, type(e), e)

    @decorators.with_dbenv()
    def run_daemon(self, *args):
        """
        Run a daemon instance in this in the current interpreter
        """
        from aiida.daemon.runner import start_daemon
        start_daemon()

    def run_listproperties(self, *args):
        """
        List all found global AiiDA properties.
        """
        import argparse

        from aiida.common.setup import (
            _property_table, exists_property, get_property)

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='List all custom properties stored in the user configuration file.')
        parser.add_argument('-a', '--all',
                            dest='all', action='store_true',
                            help="Show all properties, even if not explicitly defined, if they "
                                 "have a default value.")
        parser.set_defaults(all=False)
        parsed_args = parser.parse_args(args)

        show_all = parsed_args.all

        for prop in sorted(_property_table.keys()):
            try:
                # To enforce the generation of an exception, even if
                # there is a default value
                if show_all or exists_property(prop):
                    val = get_property(prop)
                    print "{} = {}".format(prop, val)
            except KeyError:
                pass

    def run_getproperty(self, *args):
        """
        Get a global AiiDA property from the config file in .aiida.
        """
        from aiida.common.setup import get_property

        if len(args) != 1:
            print >> sys.stderr, ("usage: {} PROPERTYNAME".format(
                self.get_full_command_name()))
            sys.exit()

        try:
            value = get_property(args[0])
        except Exception as e:
            print >> sys.stderr, ("{} while getting the "
                                  "property: {}".format(type(e).__name__, e.message))
            sys.exit(1)
        print "{}".format(value)

    def run_delproperty(self, *args):
        """
        Delete a global AiiDA property from the config file in .aiida.
        """
        from aiida.common.setup import del_property

        if len(args) != 1:
            print >> sys.stderr, ("usage: {} PROPERTYNAME".format(
                self.get_full_command_name()))
            sys.exit()

        try:
            del_property(args[0])
        except KeyError:
            print >> sys.stderr, ("No such property '{}' in the config "
                                  "file.".format(args[0]))
            sys.exit(1)

        except Exception as e:
            print >> sys.stderr, ("{} while getting the "
                                  "property: {}".format(type(e).__name__, e.message))
            sys.exit(1)

        print "Property '{}' successfully deleted.".format(args[0])

    def run_setproperty(self, *args):
        """
        Define a global AiiDA property in the config file in .aiida.

        Only properties in the _property_table of aiida.common.setup can
        be modified.
        """
        from aiida.common.setup import set_property

        if len(args) != 2:
            print >> sys.stderr, ("usage: {} PROPERTYNAME PROPERTYVALUE".format(
                self.get_full_command_name()))
            sys.exit()

        try:
            set_property(args[0], args[1])
        except Exception as e:
            print >> sys.stderr, ("{} while storing the "
                                  "property: {}".format(type(e).__name__, e.message))
            sys.exit(1)

    def run_tests(self, *args):
        import unittest
        from aiida.backends import settings
        from aiida.backends.testbase import run_aiida_db_tests
        from aiida.backends.testbase import check_if_tests_can_run
        from aiida import settings

        settings.TESTING_MODE = True

        # For final summary
        test_failures = []
        test_errors = []
        test_skipped = []
        tot_num_tests = 0

        db_test_list = []
        test_folders = []
        do_db = False
        if args:
            for arg in args:
                if arg in self.allowed_test_folders:
                    dbtests = self.allowed_test_folders[arg]
                    # Anything that has been added is a DB test
                    if dbtests is not None:
                        do_db = True
                        for dbtest in dbtests:
                            db_test_list.append(dbtest)
                    else:
                        test_folders.append(arg)
                else:
                    print >> sys.stderr, (
                        "{} is not a valid test. "
                        "Allowed test folders are:".format(arg))
                    print >> sys.stderr, "\n".join(
                        '  * {}'.format(a) for a in sorted(
                            self.allowed_test_folders.keys()))
                    sys.exit(1)
        else:
            # Without arguments, run all tests
            do_db = True
            for k, v in self.allowed_test_folders.iteritems():
                if v is None:
                    # Non-db tests
                    test_folders.append(k)
                else:
                    # DB test
                    for dbtest in v:
                        db_test_list.append(dbtest)

        for test_folder in test_folders:
            print "v" * 75
            print ">>> Tests for module {} <<<".format(test_folder.ljust(50))
            print "^" * 75
            testsuite = unittest.defaultTestLoader.discover(
                test_folder, top_level_dir=os.path.dirname(aiida.__file__))
            test_runner = unittest.TextTestRunner()
            test_results = test_runner.run(testsuite)
            test_failures.extend(test_results.failures)
            test_errors.extend(test_results.errors)
            test_skipped.extend(test_results.skipped)
            tot_num_tests += test_results.testsRun

        if do_db:
            if not is_dbenv_loaded():
                load_dbenv()

            # Even if each test would fail if we are not in a test profile,
            # it's still better to not even run them in the case the profile
            # is not a test one.
            try:
                check_if_tests_can_run()
            except TestsNotAllowedError as e:
                print >> sys.stderr, e.message
                sys.exit(1)

            # project_dir = os.path.join(os.path.dirname(aiida.__file__), '..')
            # testsuite = unittest.defaultTestLoader.discover('test', top_level_dir=project_dir)
            # test_runner = unittest.TextTestRunner()
            # test_runner.run(testsuite)

            print "v" * 75
            print (">>> Tests for {} db application".format(settings.BACKEND))
            print "^" * 75
            db_results = run_aiida_db_tests(db_test_list)
            test_skipped.extend(db_results.skipped)
            test_failures.extend(db_results.failures)
            test_errors.extend(db_results.errors)
            tot_num_tests += db_results.testsRun

        print "Final summary of the run of tests:"
        print "* Tests skipped: {}".format(len(test_skipped))
        if test_skipped:
            print "  Reasons for skipping:"
            for reason in sorted(set([_[1] for _ in test_skipped])):
                print "  - {}".format(reason)

        print "* Tests run:     {}".format(tot_num_tests)
        # This count is wrong, sometimes a test can both error and fail
        # apparently, and you can get negative numbers...
        #print "* Tests OK:      {}".format(tot_num_tests - len(test_errors) - len(test_failures))
        print "* Tests failed:  {}".format(len(test_failures))
        print "* Tests errored: {}".format(len(test_errors))



        # If there was any failure report it with the
        # right exit code
        if test_failures or test_errors:
            sys.exit(len(test_failures) + len(test_errors))

    def complete_tests(self, subargs_idx, subargs):
        """
        I complete with subargs that were not used yet.
        """
        # I remove the one on which I am, so if I wrote all of it but
        # did not press space, it will get completed
        other_subargs = subargs[:subargs_idx] + subargs[subargs_idx + 1:]
        # I create a list of the tests that are not already written on the
        # command line
        remaining_tests = (
            set(self.allowed_test_folders) - set(other_subargs))

        return " ".join(sorted(remaining_tests))


    def get_querydict_from_keyvalue(self, key, separator_filter, value):
        from aiida.orm import (Code, Data, Calculation,
                               DataFactory, CalculationFactory)
        from aiida.common.exceptions import MissingPluginError
        import re

        item, sep, subproperty = key.partition('.')

        the_type, realvalue = re.match("^(\([^\)]+\))?(.*)$", value).groups()
        value = realvalue

        if the_type is None:
            # default: string
            the_type = "(t)"

        if the_type == '(t)':
            cast_value = value
            attr_field = "tval"
        elif the_type == '(i)':
            cast_value = int(value)
            attr_field = "ival"
        elif the_type == '(f)':
            cast_value = float(value)
            attr_field = "fval"
        else:
            raise ValueError("For the moment only (t), (i) and (f) fields accepted. Got '{}' instead".format(the_type))

        do_recursive = False

        if separator_filter:
            sep_filter_string = "__{}".format(separator_filter)
        else:
            sep_filter_string = ""

        if item == 'i':  # input
            keystring = "inputs"
            do_recursive = True
        elif item == 'o':  # input
            keystring = "outputs"
            do_recursive = True
        elif item == 'p':  # input
            keystring = "parents"
            do_recursive = True
        elif item == 'c':  # input
            keystring = "children"
            do_recursive = True
        elif item == 'a':  # input
            keystring = "dbattributes"
            if '*' in subproperty:
                if sep_filter_string:
                    raise ValueError("Only = allowed if * patterns in the attributes are used")
                regex_key = re.escape(subproperty)
                # I look for non-separators, i.e. the * only covers
                # at a single level, and does not go deeper in levels
                regex_key = regex_key.replace(r'\*', r'[^\.]+')
                # Match the full string
                return ({keystring + "__key__regex": r'^{}$'.format(regex_key),
                         keystring + "__{}".format(attr_field): cast_value},
                        [keystring + "__key", keystring +
                         "__{}".format(attr_field)]
                )
            else:
                # Standard query
                return ({keystring + "__key": subproperty,
                         keystring + "__{}".format(attr_field)
                         + sep_filter_string: cast_value},
                        [keystring + "__key", keystring
                         + "__{}".format(attr_field)]
                )

        elif item == 'e':  # input
            keystring = "dbextras"

            if '*' in subproperty:
                import re

                if sep_filter_string:
                    raise ValueError("Only = allowed if * patterns in the extras are used")
                regex_key = re.escape(subproperty)
                # I look for non-separators, i.e. the * only covers
                # at a single level, and does not go deeper in levels
                regex_key = regex_key.replace(r'\*', r'[^\.]+')
                return ({keystring + "__key__regex": r'^{}$'.format(regex_key),
                         keystring + "__ival": value},
                        [keystring + "__key", keystring + "__ival"],
                )

            else:
                # Standard query
                return ({keystring + "__key": subproperty,
                         keystring + "__ival" + sep_filter_string: value},
                        [keystring + "__key", keystring + "__ival"]
                )


        elif item == 't' or item == 'type':
            if sep_filter_string:
                raise ValueError("Only = allowed for type")
            if subproperty:
                raise ValueError("Cannot pass subproperties to type")

            if value == 'code':
                DataClass = Code
            elif value == 'node':
                return ({}, [])
            elif value == 'calc' or value == 'calculation':
                DataClass = Calculation
            elif value == 'data':
                DataClass = Data
            else:
                try:
                    DataClass = DataFactory(value)
                except MissingPluginError:
                    try:
                        DataClass = CalculationFactory(value)
                    except MissingPluginError:
                        raise ValueError("Unknown node type '{}'".format(value))

            # Startswith so to get also subclasses
            return ({"type__startswith": DataClass._query_type_string},
                    ["type"])

        else:
            # this is a 'direct' attribute: just create the query
            return ({key + sep_filter_string: value},  # all, including the dots
                    [key])

        if do_recursive:
            if sep:
                tempdict, templist = self.get_querydict_from_keyvalue(
                    subproperty,
                    separator_filter,
                    value)
                # In this case the filter is added by the recursive function
                return ({"{}__{}".format(keystring, k): v
                         for k, v in tempdict.iteritems()},
                        ["{}__{}".format(keystring, k) for k in templist] +
                        ["{}__pk".format(keystring),
                         "{}__type".format(keystring),
                         "{}__label".format(keystring)],
                )
            else:
                return ({keystring + sep_filter_string: value},
                        [keystring] +
                        ["{}__pk".format(keystring),
                         "{}__type".format(keystring),
                         "{}__label".format(keystring)],
                )

        raise NotImplementedError("Should not be here...")


    # To be put in the right order! For instance,
    # You want to put ~= before =, because this has
    # to be checked first
    # First element of each tuple: string used as separator
    # Second element of each tuple: django filter to apply
    separators = [
        ("~=", "iexact"),
        (">=", "gte"),
        ("<=", "lte"),
        (">", "gt"),
        ("<", "lt"),
        ("=", ""),
    ]
    # TO ADD: startswith, istartswith, endswith, iendswith, ymdHMS, isnull, in, contains,

    # TO ADD: support for other types


    def parse_arg(self, arg):
        # pieces = arg.split("=")
        #        if len(pieces) != 2:
        #            raise ValueError("Each option must be in the key=value format")
        #        key = pieces[0]
        #        value = pieces[1]

        #import re
        #sep_regex = "|".join([re.escape(k) for k in self.separators.keys()])
        #regex = re.match(r'(.+)(' + sep_regex + r')(.+)',arg)
        #        key, sep, value = regex.groups()

        for s, django_filter in self.separators:
            key, sep, value = arg.partition(s)
            if sep:
                filter_to_apply = django_filter
                break
        if not sep:
            raise ValueError("No separator provided!")

        querydict = self.get_querydict_from_keyvalue(key,
                                                     filter_to_apply,
                                                     value)

        return querydict


    def run_query(self, *args):
        load_dbenv()
        from django.db.models import Q
        from aiida.backends.djsite.db.models import DbNode

        # django_query = Q()
        django_query = DbNode.objects.filter()

        try:
            # NOT SURE THIS IS THE RIGHT WAY OF MANAGING NEGATION FOR
            # ATTRIBUTES!!
            queries = [(self.parse_arg(arg[1:] if arg.startswith('!') else arg), arg.startswith('!')) for arg in args]
        except ValueError as e:
            print "ERROR in format!"
            print e.message
            sys.exit(1)

        all_values = ['pk']

        print "*** DEBUG: QUERY ***" + "*" * 52
        for (q, v), n in queries:
            print "* {}{}".format("[NEG]" if n else "", q)
        print "*" * 72

        for (query, values), negate in queries:

            all_values += list(values)
            if negate:
                if len(query) == 1:
                    #                    django_query = django_query & ~Q(**query)
                    django_query = django_query.filter(~Q(**query))
                elif len(query) == 2:

                    raise NotImplementedError(
                        "The current implementation does not work for negation of attributes. See comments in source code.")

                    ## NOT THE RIGHT WAY TO MANAGE NEGATIONS!! SEE AT THE
                    ## END OF THE FILE
                    temp = [(k, v) for k, v in query.iteritems()]
                    if temp[0][0].endswith("__key"):
                        #                        django_query = django_query & (
                        #                            Q(~Q(**{temp[1][0]: temp[1][1]}),
                        #                               **{temp[0][0]: temp[0][1]})
                        django_query = django_query.filter(~Q(**{temp[1][0]: temp[1][1]}),
                                                           **{temp[0][0]: temp[0][1]})
                    elif temp[1][0].endswith("__key"):
                        django_query = django_query.filter(~Q(**{temp[0][0]: temp[0][1]}),
                                                           **{temp[1][0]: temp[1][1]})
                    else:
                        raise NotImplementedError("Should not be here... no key search?")
                else:
                    raise NotImplementedError("Should not be here...")
            else:
                django_query = django_query.filter(Q(**query))

        res = django_query.distinct().order_by('pk')
        res = res.values_list(*all_values)
        print res.query
        print "{} matching nodes found ({} removing duplicates).".format(len(res),
                                                                         len(set([_[0] for _ in res])))
        #        for node in res:
        #            print "* {}".format(str(node))
        #            print "* {} ({}){}".format(
        #                node.pk, node.get_aiida_class().__class__.__name__,
        #                " [{}]".format(node.label) if node.label else "")
        for node in res:
            print "* {}".format(node[0])
            for p, v in zip(all_values[1:], node[1:]):
                print "  `-> {} = {}".format(p, v)


    def run_play(self, *args):
        """
        Open a browser and play the Aida triumphal march by Giuseppe Verdi
        """
        import webbrowser

        webbrowser.open_new('http://upload.wikimedia.org/wikipedia/commons/3/32/Triumphal_March_from_Aida.ogg')

# In [11]: attr_res = DbAttribute.objects.filter(Q(key='cell.atoms'), ~Q(ival__gt=7))
#
#In [12]: dbres = DbNode.objects.filter(outputs__dbattributes__in=attr_res).distinct()
#
#In [13]: print dbres.query
#SELECT DISTINCT "db_dbnode"."id", "db_dbnode"."uuid", "db_dbnode"."type", "db_dbnode"."label", "db_dbnode"."description", "db_dbnode"."ctime", "db_dbnode"."mtime", "db_dbnode"."user_id", "db_dbnode"."dbcomputer_id", "db_dbnode"."nodeversion", "db_dbnode"."lastsyncedversion" FROM "db_dbnode" INNER JOIN "db_dblink" ON ("db_dbnode"."id" = "db_dblink"."input_id") INNER JOIN "db_dbnode" T3 ON ("db_dblink"."output_id" = T3."id") INNER JOIN "db_dbattribute" ON (T3."id" = "db_dbattribute"."dbnode_id") WHERE "db_dbattribute"."id" IN (SELECT U0."id" FROM "db_dbattribute" U0 WHERE (U0."key" = cell.atoms  AND NOT ((U0."ival" > 7  AND U0."ival" IS NOT NULL))))

## OR, if doing
#dbres = DbNode.objects.filter(dbattributes__in=attr_res).distinct()
#In [18]: print dbres.query
#SELECT DISTINCT "db_dbnode"."id", "db_dbnode"."uuid", "db_dbnode"."type", "db_dbnode"."label", "db_dbnode"."description", "db_dbnode"."ctime", "db_dbnode"."mtime", "db_dbnode"."user_id", "db_dbnode"."dbcomputer_id", "db_dbnode"."nodeversion", "db_dbnode"."lastsyncedversion" FROM "db_dbnode" INNER JOIN "db_dbattribute" ON ("db_dbnode"."id" = "db_dbattribute"."dbnode_id") WHERE "db_dbattribute"."id" IN (SELECT U0."id" FROM "db_dbattribute" U0 WHERE (U0."key" = cell.atoms  AND NOT ((U0."ival" > 7  AND U0."ival" IS NOT NULL))))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This allows to manage profiles from command line.
"""
import click
import sys

from aiida.cmdline.baseclass import VerdiCommandWithSubcommands
from aiida.control.postgres import Postgres


class Profile(VerdiCommandWithSubcommands):
    """
    List AiiDA profiles, and set the default profile.

    Allow to see the list of AiiDA profiles, and to set the default profile
    (the to be used by any verdi command when no '-p' option is given).
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        self.valid_subcommands = {
            'setdefault': (self.profile_setdefault, self.complete_profiles),
            'list': (self.profile_list, self.complete_none),
            'delete': (self.profile_delete, self.complete_profiles),
        }

    def complete_profiles(self, subargs_idx, subargs):
        from aiida.common.setup import get_profiles_list

        if subargs_idx == 0:
            return "\n".join(get_profiles_list())
        else:
            return ""

    def profile_setdefault(self, *args):
        from aiida.common.setup import set_default_profile

        if len(args) != 1:
            print >> sys.stderr, ("Please specify the profile to be set as the default")
            sys.exit(1)

        profile = args[0]
        # set default DB profiles
        set_default_profile(profile, force_rewrite=True)


    def profile_list(self, *args):
        from aiida.backends import settings
        from aiida.common.setup import get_profiles_list, get_default_profile, AIIDA_CONFIG_FOLDER
        from aiida.common.exceptions import ConfigurationError

        print('Configuration folder: {}'.format(AIIDA_CONFIG_FOLDER))

        try:
            default_profile = get_default_profile()
        except ConfigurationError as e:
            err_msg = (
                "Stopping: {}\n"
                "Note: if no configuration file was found, it means that you have not run\n" 
            "'verdi setup' yet to configure at least one AiiDA profile.".format(e.message))
            click.echo(err_msg, err=True)
            sys.exit(1)

        use_colors = True
        if args:
            try:
                if len(args) != 1:
                    raise ValueError
                if args[0] != '--no-color':
                    raise ValueError
                use_colors = False
            except ValueError:
                print >> sys.stderr, ('You can pass only one further argument, '
                                      '--no-color, to show the results without colors')
                sys.exit(1)

        if default_profile is None:
            print >> sys.stderr, "### No default profile configured yet, run 'verdi install'! ###"
            return
        else:
            print >> sys.stderr, 'The default profile is highlighted and marked by the * symbol'

        for profile in get_profiles_list():
            color_id = 39  # Default foreground color
            if profile == default_profile:
                symbol = '*'
                color_id = 32
            else:
                symbol = ' '

            if use_colors:
                start_color = '\x1b[{}m'.format(color_id)
                end_color = '\x1b[0m'
                bold_sequence = '\x1b[1;{}m'.format(color_id)
                nobold_sequence = '\x1b[0;{}m'.format(color_id)
            else:
                start_color = ''
                end_color = ''
                bold_sequence = ''
                nobold_sequence = ''

            print '{}{} {}{} {}{}'.format(
                start_color, symbol, nobold_sequence, profile, nobold_sequence, end_color)

    def profile_delete(self, *args):
        """ Deletes profile

        Asks whether to delete associated database and associated database
        user.

        Specify argument '--force' to skip any questions warning about loss of
        data.
        """
        from aiida.common.setup import get_or_create_config, update_config
        import os.path
        from urlparse import urlparse

        args = list(args)
        if '--force' in args:
            force = True
            args.remove('--force')
        else:
            force = False

        confs = get_or_create_config()
        profiles = confs.get('profiles',{})
        users = [ profiles[name].get('AIIDADB_USER', '') for name in profiles.keys()]

        profiles_to_delete = args
        for profile_to_delete in profiles_to_delete:
            try:
                profile = profiles[profile_to_delete]
            except KeyError:
                print("Profile '{}' does not exist".format(profile_to_delete))
                continue

            postgres = Postgres(port=profile.get('AIIDADB_PORT'), interactive=True, quiet=False)
            postgres.determine_setup()
            print postgres.dbinfo

            db_name = profile.get('AIIDADB_NAME', '')
            if not postgres.db_exists(db_name):
                print("Associated database '{}' does not exist.".format(db_name))
            elif force or click.confirm("Delete associated database '{}'?\n" \
                             "WARNING: All data will be lost.".format(db_name)):
                print("Deleting database '{}'.".format(db_name))
                postgres.drop_db(db_name)

            user = profile.get('AIIDADB_USER', '')
            if not postgres.dbuser_exists(user):
                print("Associated database user '{}' does not exist.".format(user))
            elif users.count(user) > 1:
                print("Associated database user '{}' is used by other profiles "\
                      "and will not be deleted.".format(user))
            elif force or click.confirm("Delete database user '{}'?".format(user)):
                print("Deleting user '{}'.".format(user))
                postgres.drop_dbuser(user)

            repo_uri = profile.get('AIIDADB_REPOSITORY_URI','')
            repo_path = urlparse(repo_uri).path
            repo_path = os.path.expanduser(repo_path)
            if not os.path.isabs(repo_path):
                print("Associated file repository '{}' does not exist."\
                      .format(repo_path))
            elif not os.path.isdir(repo_path):
                print("Associated file repository '{}' is not a directory."\
                       .format(repo_path))
            elif force or click.confirm("Delete associated file repository '{}'?\n" \
                               "WARNING: All data will be lost.".format(repo_path)):
                print("Deleting directory '{}'.".format(repo_path))
                import shutil
                shutil.rmtree(repo_path)

            if force or click.confirm("Delete configuration for profile '{}'?\n" \
                             "WARNING: Permanently removes profile from the list of AiiDA profiles."\
                             .format(profile_to_delete)):
                print("Deleting configuration for profile '{}'.".format(profile_to_delete))
                del profiles[profile_to_delete]
                update_config(confs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.common.exceptions import InternalError
from abc import ABCMeta, abstractmethod


class AiidaTestImplementation(object):
    """
    For each implementation, define what to do at setUp and tearDown.

    Each subclass must reimplement two *standard* methods (i.e., *not* classmethods), called
    respectively ``setUpClass_method`` and ``tearDownClass_method``.
    It is also required to implement setUp_method and tearDown_method to be run for each single test
    They can set local properties (e.g. ``self.xxx = yyy``) but remember that ``xxx``
    is not visible to the upper (calling) Test class.

    Moreover, it is required that they define in the setUpClass_method the two properties:

    - ``self.computer`` that must be a Computer object
    - ``self.user_email`` that must be a string

    These two are then exposed by the ``self.get_computer()`` and ``self.get_user_email()``
    methods.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def setUpClass_method(self):
        """
        This class prepares the database (cleans it up and installs some basic entries).
        You have also to set a self.computer and a self.user_email as explained in the docstring of the
        AiidaTestImplemention docstring.
        """
        pass

    @abstractmethod
    def setUp_method(self):
        pass

    @abstractmethod
    def tearDown_method(self):
        pass

    @abstractmethod
    def tearDownClass_method(self):
        """
        This class implements the tear down methods (e.g. cleans up the DB).
        """
        pass

    @abstractmethod
    def clean_db(self):
        """
        This method implements the logic to fully clean the DB.
        """
        pass

    @abstractmethod
    def insert_data(self):
        """
        This method inserts default data into the database.
        """
        pass

    def get_computer(self):
        """
        A ORM Computer object present in the DB
        """
        try:
            return self.computer
        except AttributeError:
            raise InternalError("The AiiDA Test implementation should define a self.computer in the setUpClass_method")

    def get_user_email(self):
        """
        A string with the email of the User
        """
        try:
            return self.user_email
        except AttributeError:
            raise InternalError("The AiiDA Test implementation should define a self.computer in the setUpClass_method")




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""" Resources for REST API """

from urllib import unquote

from flask import request, make_response
from flask_restful import Resource

from aiida.restapi.common.utils import Utils


# pylint: disable=missing-docstring,fixme
class ServerInfo(Resource):

    def __init__(self, **kwargs):
        # Configure utils
        utils_conf_keys = ('PREFIX', 'PERPAGE_DEFAULT', 'LIMIT_DEFAULT')
        self.utils_confs = {k: kwargs[k] for k in utils_conf_keys if k in kwargs}
        self.utils = Utils(**self.utils_confs)

    def get(self):
        """
        It returns the general info about the REST API
        :return: returns current AiiDA version defined in aiida/__init__.py
        """

        ## Decode url parts
        path = unquote(request.path)
        query_string = unquote(request.query_string)
        url = unquote(request.url)
        url_root = unquote(request.url_root)

        pathlist = self.utils.split_path(self.utils.strip_prefix(path))

        if len(pathlist) > 1:
            resource_type = pathlist.pop(1)
        else:
            resource_type = "info"

        response = {}

        import aiida.restapi.common.config as conf
        from aiida import __version__

        if resource_type == "info":
            response = []

            # Add Rest API version
            response.append("REST API version: " + conf.PREFIX.split("/")[-1])

            # Add Rest API prefix
            response.append("REST API Prefix: " + conf.PREFIX)

            # Add AiiDA version
            response.append("AiiDA==" + __version__)

        elif resource_type == "endpoints":

            from aiida.restapi.common.utils import list_routes
            response["available_endpoints"] = list_routes()

        headers = self.utils.build_headers(url=request.url, total_count=1)

        ## Build response and return it
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=path,
            query_string=query_string,
            resource_type="Info",
            data=response)
        return self.utils.build_response(status=200, headers=headers, data=data)


## TODO add the caching support. I cache total count, results, and possibly
# set_query
class BaseResource(Resource):
    """
    Each derived class will instantiate a different type of translator.
    This is the only difference in the classes.
    """

    def __init__(self, **kwargs):

        self.trans = None

        # Flag to tell the path parser whether to expect a pk or a uuid pattern
        self.parse_pk_uuid = None

        # Configure utils
        utils_conf_keys = ('PREFIX', 'PERPAGE_DEFAULT', 'LIMIT_DEFAULT')
        self.utils_confs = {k: kwargs[k] for k in utils_conf_keys if k in kwargs}
        self.utils = Utils(**self.utils_confs)
        self.method_decorators = {'get': kwargs.get('get_decorators', [])}

    # pylint: disable=too-many-locals,redefined-builtin,invalid-name
    def get(self, id=None, page=None):
        """
        Get method for the Computer resource
        :return:
        """

        ## Decode url parts
        path = unquote(request.path)
        query_string = unquote(request.query_string)
        url = unquote(request.url)
        url_root = unquote(request.url_root)

        ## Parse request
        (resource_type, page, id, query_type) = self.utils.parse_path(path, parse_pk_uuid=self.parse_pk_uuid)
        (limit, offset, perpage, orderby, filters, _alist, _nalist, _elist, _nelist, _downloadformat, _visformat,
         _filename, _rtype) = self.utils.parse_query_string(query_string)

        ## Validate request
        self.utils.validate_request(
            limit=limit,
            offset=offset,
            perpage=perpage,
            page=page,
            query_type=query_type,
            is_querystring_defined=(bool(query_string)))

        ## Treat the schema case which does not imply access to the DataBase
        if query_type == 'schema':

            ## Retrieve the schema
            results = self.trans.get_schema()
            ## Build response and return it
            headers = self.utils.build_headers(url=request.url, total_count=1)

        else:
            ## Set the query, and initialize qb object
            self.trans.set_query(filters=filters, orders=orderby, id=id)

            ## Count results
            total_count = self.trans.get_total_count()

            ## Pagination (if required)
            if page is not None:
                (limit, offset, rel_pages) = self.utils.paginate(page, perpage, total_count)
                self.trans.set_limit_offset(limit=limit, offset=offset)
                headers = self.utils.build_headers(rel_pages=rel_pages, url=request.url, total_count=total_count)
            else:
                self.trans.set_limit_offset(limit=limit, offset=offset)
                headers = self.utils.build_headers(url=request.url, total_count=total_count)

            ## Retrieve results
            results = self.trans.get_results()

        ## Build response and return it
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=request.path,
            id=id,
            query_string=request.query_string,
            resource_type=resource_type,
            data=results)
        return self.utils.build_response(status=200, headers=headers, data=data)


class Node(Resource):
    """
    Differs from BaseResource in trans.set_query() mostly because it takes
    query_type as an input and the presence of additional result types like "tree"
    """

    def __init__(self, **kwargs):

        # Set translator
        from aiida.restapi.translator.node import NodeTranslator
        self.trans = NodeTranslator(**kwargs)

        from aiida.orm import Node as tNode
        self.tclass = tNode

        # Parse a uuid pattern in the URL path (not a pk)
        self.parse_pk_uuid = 'uuid'

        # Configure utils
        utils_conf_keys = ('PREFIX', 'PERPAGE_DEFAULT', 'LIMIT_DEFAULT')
        self.utils_confs = {k: kwargs[k] for k in utils_conf_keys if k in kwargs}
        self.utils = Utils(**self.utils_confs)
        self.method_decorators = {'get': kwargs.get('get_decorators', [])}

    #pylint: disable=too-many-locals,too-many-statements
    #pylint: disable=redefined-builtin,invalid-name,too-many-branches
    def get(self, id=None, page=None):
        """
        Get method for the Node resource.
        :return:
        """

        ## Decode url parts
        path = unquote(request.path)
        query_string = unquote(request.query_string)
        url = unquote(request.url)
        url_root = unquote(request.url_root)

        ## Parse request
        (resource_type, page, id, query_type) = self.utils.parse_path(path, parse_pk_uuid=self.parse_pk_uuid)

        (limit, offset, perpage, orderby, filters, alist, nalist, elist, nelist, downloadformat, visformat, filename,
         rtype) = self.utils.parse_query_string(query_string)

        ## Validate request
        self.utils.validate_request(
            limit=limit,
            offset=offset,
            perpage=perpage,
            page=page,
            query_type=query_type,
            is_querystring_defined=(bool(query_string)))

        ## Treat the schema case which does not imply access to the DataBase
        if query_type == 'schema':

            ## Retrieve the schema
            results = self.trans.get_schema()

            ## Build response and return it
            headers = self.utils.build_headers(url=request.url, total_count=1)

        ## Treat the statistics
        elif query_type == "statistics":
            (limit, offset, perpage, orderby, filters, alist, nalist, elist, nelist, downloadformat, visformat,
             filename, rtype) = self.utils.parse_query_string(query_string)
            headers = self.utils.build_headers(url=request.url, total_count=0)
            if filters:
                usr = filters["user"]["=="]
            else:
                usr = None
            results = self.trans.get_statistics(usr)

        # TODO Might need to be improved
        elif query_type == "tree":
            headers = self.utils.build_headers(url=request.url, total_count=0)
            results = self.trans.get_io_tree(id)
        else:
            ## Initialize the translator
            self.trans.set_query(
                filters=filters,
                orders=orderby,
                query_type=query_type,
                id=id,
                alist=alist,
                nalist=nalist,
                elist=elist,
                nelist=nelist,
                downloadformat=downloadformat,
                visformat=visformat,
                filename=filename,
                rtype=rtype)

            ## Count results
            total_count = self.trans.get_total_count()

            ## Pagination (if required)
            if page is not None:
                (limit, offset, rel_pages) = self.utils.paginate(page, perpage, total_count)
                self.trans.set_limit_offset(limit=limit, offset=offset)

                ## Retrieve results
                results = self.trans.get_results()

                headers = self.utils.build_headers(rel_pages=rel_pages, url=request.url, total_count=total_count)
            else:

                self.trans.set_limit_offset(limit=limit, offset=offset)
                ## Retrieve results
                results = self.trans.get_results()

                if query_type == "download" and results:
                    if results["download"]["status"] == 200:
                        data = results["download"]["data"]
                        response = make_response(data)
                        response.headers['content-type'] = 'application/octet-stream'
                        response.headers['Content-Disposition'] = 'attachment; filename="{}"'.format(
                            results["download"]["filename"])
                        return response

                    else:
                        results = results["download"]["data"]

                if query_type in ["retrieved_inputs", "retrieved_outputs"] and results:
                    try:
                        status = results[query_type]["status"]
                    except KeyError:
                        status = ""
                    except TypeError:
                        status = ""

                    if status == 200:
                        data = results[query_type]["data"]
                        response = make_response(data)
                        response.headers['content-type'] = 'application/octet-stream'
                        response.headers['Content-Disposition'] = 'attachment; filename="{}"'.format(
                            results[query_type]["filename"])
                        return response

                    elif status == 500:
                        results = results[query_type]["data"]

                headers = self.utils.build_headers(url=request.url, total_count=total_count)

        ## Build response
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=path,
            id=id,
            query_string=query_string,
            resource_type=resource_type,
            data=results)

        return self.utils.build_response(status=200, headers=headers, data=data)


class Computer(BaseResource):

    def __init__(self, **kwargs):
        super(Computer, self).__init__(**kwargs)

        ## Instantiate the correspondent translator
        from aiida.restapi.translator.computer import ComputerTranslator
        self.trans = ComputerTranslator(**kwargs)

        # Set wheteher to expect a pk (integer) or a uuid pattern (string) in
        # the URL path
        self.parse_pk_uuid = "uuid"


class Group(BaseResource):

    def __init__(self, **kwargs):
        super(Group, self).__init__(**kwargs)

        from aiida.restapi.translator.group import GroupTranslator
        self.trans = GroupTranslator(**kwargs)

        self.parse_pk_uuid = 'uuid'


class User(BaseResource):

    def __init__(self, **kwargs):
        super(User, self).__init__(**kwargs)

        from aiida.restapi.translator.user import UserTranslator
        self.trans = UserTranslator(**kwargs)

        self.parse_pk_uuid = 'pk'


class Calculation(Node):

    def __init__(self, **kwargs):
        super(Calculation, self).__init__(**kwargs)

        from aiida.restapi.translator.calculation import CalculationTranslator
        self.trans = CalculationTranslator(**kwargs)
        from aiida.orm import Calculation as CalculationTclass
        self.tclass = CalculationTclass

        self.parse_pk_uuid = 'uuid'


class Code(Node):

    def __init__(self, **kwargs):
        super(Code, self).__init__(**kwargs)

        from aiida.restapi.translator.code import CodeTranslator
        self.trans = CodeTranslator(**kwargs)
        from aiida.orm import Code as CodeTclass
        self.tclass = CodeTclass

        self.parse_pk_uuid = 'uuid'


class Data(Node):

    def __init__(self, **kwargs):
        super(Data, self).__init__(**kwargs)

        from aiida.restapi.translator.data import DataTranslator
        self.trans = DataTranslator(**kwargs)
        from aiida.orm import Data as DataTclass
        self.tclass = DataTclass

        self.parse_pk_uuid = 'uuid'


class StructureData(Data):

    def __init__(self, **kwargs):

        super(StructureData, self).__init__(**kwargs)

        from aiida.restapi.translator.data.structure import \
            StructureDataTranslator
        self.trans = StructureDataTranslator(**kwargs)
        from aiida.orm.data.structure import StructureData as StructureDataTclass
        self.tclass = StructureDataTclass

        self.parse_pk_uuid = 'uuid'


class KpointsData(Data):

    def __init__(self, **kwargs):
        super(KpointsData, self).__init__(**kwargs)

        from aiida.restapi.translator.data.kpoints import KpointsDataTranslator
        self.trans = KpointsDataTranslator(**kwargs)
        from aiida.orm.data.array.kpoints import KpointsData as KpointsDataTclass
        self.tclass = KpointsDataTclass

        self.parse_pk_uuid = 'uuid'


class BandsData(Data):

    def __init__(self, **kwargs):
        super(BandsData, self).__init__(**kwargs)

        from aiida.restapi.translator.data.bands import \
            BandsDataTranslator
        self.trans = BandsDataTranslator(**kwargs)
        from aiida.orm.data.array.bands import BandsData as BandsDataTclass
        self.tclass = BandsDataTclass

        self.parse_pk_uuid = 'uuid'


class CifData(Data):

    def __init__(self, **kwargs):

        super(CifData, self).__init__(**kwargs)

        from aiida.restapi.translator.data.cif import \
            CifDataTranslator
        self.trans = CifDataTranslator(**kwargs)
        from aiida.orm.data.cif import CifData as CifDataTclass
        self.tclass = CifDataTclass

        self.parse_pk_uuid = 'uuid'


class UpfData(Data):

    def __init__(self, **kwargs):

        super(UpfData, self).__init__(**kwargs)

        from aiida.restapi.translator.data.upf import \
            UpfDataTranslator
        self.trans = UpfDataTranslator(**kwargs)
        from aiida.orm.data.upf import UpfData as UpfDataTclass
        self.tclass = UpfDataTclass

        self.parse_pk_uuid = 'uuid'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA specific implementation of plumpy's ProcessSpec."""
import plumpy

from aiida.work.exit_code import ExitCode, ExitCodesNamespace
from aiida.work.ports import InputPort, PortNamespace


class ProcessSpec(plumpy.ProcessSpec):
    """
    Sub classes plumpy.ProcessSpec to define the INPUT_PORT_TYPE and PORT_NAMESPACE_TYPE
    with the variants implemented in AiiDA
    """

    INPUT_PORT_TYPE = InputPort
    PORT_NAMESPACE_TYPE = PortNamespace

    def __init__(self):
        super(ProcessSpec, self).__init__()
        self._exit_codes = ExitCodesNamespace()

    @property
    def exit_codes(self):
        """
        Return the namespace of exit codes defined for this ProcessSpec

        :returns: ExitCodesNamespace of ExitCode named tuples
        """
        return self._exit_codes

    def exit_code(self, status, label, message):
        """
        Add an exit code to the ProcessSpec

        :param status: the exit status integer
        :param label: a label by which the exit code can be addressed
        :param message: a more detailed description of the exit code
        """
        if not isinstance(status, int):
            raise TypeError('status should be of integer type and not of {}'.format(type(status)))

        if status < 0:
            raise ValueError('status should be a positive integer, received {}'.format(type(status)))

        if not isinstance(label, basestring):
            raise TypeError('label should be of basestring type and not of {}'.format(type(label)))

        if not isinstance(message, basestring):
            raise TypeError('message should be of basestring type and not of {}'.format(type(message)))

        self._exit_codes[label] = ExitCode(status, message)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import absolute_import

from aiida.backends.settings import BACKEND
from aiida.common.exceptions import ConfigurationError
from aiida.backends.profile import BACKEND_DJANGO, BACKEND_SQLA

from aiida.orm.implementation.general.calculation.job import _input_subfolder
from aiida.orm.implementation.general.calculation.job import JobCalculationExitStatus


if BACKEND == BACKEND_SQLA:
    from aiida.orm.implementation.sqlalchemy.calculation import Calculation
    from aiida.orm.implementation.sqlalchemy.calculation.job import JobCalculation

elif BACKEND == BACKEND_DJANGO:
    from aiida.orm.implementation.django.calculation import Calculation
    from aiida.orm.implementation.django.calculation.job import JobCalculation

else:
    raise ConfigurationError("Invalid settings.BACKEND: {}".format(
                BACKEND))




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import json

from copy import copy

from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.session import make_transient
from sqlalchemy.orm.attributes import flag_modified

from aiida.backends.sqlalchemy.models.computer import DbComputer
from aiida.common.exceptions import (NotExistent, ConfigurationError,
                                     InvalidOperation)
from aiida.orm.implementation.general.computer import AbstractComputer, Util as ComputerUtil
from aiida.common.lang import override



class Computer(AbstractComputer):

    @property
    def uuid(self):
        return unicode(self._dbcomputer.uuid)

    @property
    def pk(self):
        return self._dbcomputer.id

    @property
    def id(self):
        return self._dbcomputer.id

    def __init__(self, **kwargs):
        super(Computer, self).__init__()

        uuid = kwargs.pop('uuid', None)
        if uuid is not None:
            if kwargs:
                raise ValueError("If you pass a uuid, you cannot pass any "
                                 "further parameter")

            dbcomputer = DbComputer.query.filter_by(uuid=uuid).first()
            if not dbcomputer:
                raise NotExistent("No entry with UUID={} found".format(uuid))

            self._dbcomputer = dbcomputer

        else:
            if 'dbcomputer' in kwargs:
                dbcomputer = kwargs.pop('dbcomputer')
                if not (isinstance(dbcomputer, DbComputer)):
                    raise TypeError("dbcomputer must be of type DbComputer")
                self._dbcomputer = dbcomputer

                if kwargs:
                    raise ValueError("If you pass a dbcomputer parameter, "
                                     "you cannot pass any further parameter")
            else:
                self._dbcomputer = DbComputer()

            # Set all remaining parameters, stop if unknown
            self.set(**kwargs)

    def set(self, **kwargs):

        is_modified = False

        for key, val in kwargs.iteritems():
            if hasattr(self._dbcomputer, key):
                setattr(self._dbcomputer, key, val)
            else:
                self._dbcomputer._metadata[key] = val
                is_modified = True

        if is_modified:
            flag_modified(self._dbcomputer, "_metadata")

    @classmethod
    def list_names(cls):
        from aiida.backends.sqlalchemy import get_scoped_session
        session = get_scoped_session()
        return session.query(DbComputer.name).all()

    @property
    def full_text_info(self):
        ret_lines = []
        ret_lines.append("Computer name:     {}".format(self.name))
        ret_lines.append(" * PK:             {}".format(self.pk))
        ret_lines.append(" * UUID:           {}".format(self.uuid))
        ret_lines.append(" * Description:    {}".format(self.description))
        ret_lines.append(" * Hostname:       {}".format(self.hostname))
        ret_lines.append(" * Enabled:        {}".format("True" if self.is_enabled() else "False"))
        ret_lines.append(" * Transport type: {}".format(self.get_transport_type()))
        ret_lines.append(" * Scheduler type: {}".format(self.get_scheduler_type()))
        ret_lines.append(" * Work directory: {}".format(self.get_workdir()))
        ret_lines.append(" * Shebang:        {}".format(self.get_shebang()))
        ret_lines.append(" * mpirun command: {}".format(" ".join(
            self.get_mpirun_command())))
        def_cpus_machine = self.get_default_mpiprocs_per_machine()

        if def_cpus_machine is not None:
            ret_lines.append(" * Default number of cpus per machine: {}".format(
                def_cpus_machine))
        ret_lines.append(" * Used by:        {} nodes".format(
            len(self.dbcomputer.dbnodes)))

        ret_lines.append(" * prepend text:")

        if self.get_prepend_text().strip():
            for l in self.get_prepend_text().split('\n'):
                ret_lines.append("   {}".format(l))
        else:
            ret_lines.append("   # No prepend text.")

        ret_lines.append(" * append text:")

        if self.get_append_text().strip():
            for l in self.get_append_text().split('\n'):
                ret_lines.append("   {}".format(l))
        else:
            ret_lines.append("   # No append text.")

        return "\n".join(ret_lines)

    @property
    def to_be_stored(self):
        return self._dbcomputer.id is None

    @classmethod
    def get(cls, computer):
        return cls(dbcomputer=DbComputer.get_dbcomputer(computer))

    @override
    def copy(self):
        from aiida.backends.sqlalchemy import get_scoped_session
        session = get_scoped_session()

        if self.to_be_stored:
            raise InvalidOperation("You can copy a computer only after having stored it")

        newdbcomputer = copy(self.dbcomputer)
        make_transient(newdbcomputer)
        session.add(newdbcomputer)

        newobject = self.__class__(dbcomputer=newdbcomputer)

        return newobject

    @property
    def dbcomputer(self):
        return self._dbcomputer

    def store(self):
        self.validate()

        try:
            self.dbcomputer.save(commit=True)
        except SQLAlchemyError:
            raise ValueError(
                "Integrity error, probably the hostname already exists in the"
                " DB")

        return self

    @property
    def name(self):
        return self.dbcomputer.name

    @property
    def description(self):
        return self.dbcomputer.description

    @property
    def hostname(self):
        return self.dbcomputer.hostname

    def _get_metadata(self):
        return self.dbcomputer._metadata

    def _set_metadata(self, metadata_dict):
        self.dbcomputer._metadata = metadata_dict
        flag_modified(self.dbcomputer, "_metadata")
        if not self.to_be_stored:
            self.dbcomputer.save()

    def get_transport_params(self):
        """
        Return transport params stored in dbcomputer instance
        """
        return self.dbcomputer.transport_params

    def set_transport_params(self, val):
        # if self.to_be_stored:
        try:
            json.dumps(val) # Check if json compatible
            self.dbcomputer.transport_params = val
        except ValueError:
            raise ValueError("The set of transport_params are not JSON-able")
        if not self.to_be_stored:
            self.dbcomputer.save()

        #        else:
        #            raise ModificationNotAllowed("Cannot set a property after having stored the entry")

    def get_workdir(self):
        try:
            return self.dbcomputer.get_workdir()
        except ConfigurationError:
            # This happens the first time: I provide a reasonable default value
            return "/scratch/{username}/aiida_run/"

    def set_workdir(self, val):
        pass
        # if self.to_be_stored:
        metadata = self._get_metadata()
        metadata['workdir'] = val
        self._set_metadata(metadata)

    def get_shebang(self):
        try:
            return self.dbcomputer.get_shebang()
        except ConfigurationError:
            # This happens the first time: I provide a reasonable default value
            return "#!/bin/bash"

    def get_name(self):
        return self.dbcomputer.name

    def set_name(self, val):
        self.dbcomputer.name = val
        if not self.to_be_stored:
            self.dbcomputer.save()

    def get_hostname(self):
        return self.dbcomputer.hostname

    def set_hostname(self, val):
        self.dbcomputer.hostname = val
        if not self.to_be_stored:
            self.dbcomputer.save()

    def get_description(self):
        return self.dbcomputer.description

    def set_description(self, val):
        self.dbcomputer.description = val
        if not self.to_be_stored:
            self.dbcomputer.save()

    def get_calculations_on_computer(self):
        from aiida.backends.sqlalchemy.models.node import DbNode
        return DbNode.query.filter(
            DbNode.dbcomputer_id == self.dbcomputer.id,
            DbNode.type.like("calculation%")).all()

    def is_enabled(self):
        return self.dbcomputer.enabled

    def set_enabled_state(self, enabled):
        self.dbcomputer.enabled = enabled
        if not self.to_be_stored:
            self.dbcomputer.save()

    def get_scheduler_type(self):
        return self.dbcomputer.scheduler_type

    def set_scheduler_type(self, val):

        self.dbcomputer.scheduler_type = val
        if not self.to_be_stored:
            self.dbcomputer.save()

    def get_transport_type(self):
        return self.dbcomputer.transport_type

    def set_transport_type(self, val):
        self.dbcomputer.transport_type = val
        if not self.to_be_stored:
            self.dbcomputer.save()


class Util(ComputerUtil):
    @override
    def delete_computer(self, pk):
        """
        Delete the computer with the given pk.
        :param pk: The computer pk.
        """
        import aiida.backends.sqlalchemy
        try:
            session = aiida.backends.sqlalchemy.get_scoped_session()
            session.query(DbComputer).get(pk).delete()
            session.commit()
        except SQLAlchemyError as e:
            raise InvalidOperation("Unable to delete the requested computer: it is possible that there "
                                   "is at least one node using this computer (original message: {})".format(
                e.message
            ))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import click

from aiida.cmdline.utils.pluginable import Pluginable


def click_subcmd_complete(cmd_group):
    """Create a subcommand completion function for a click command group."""
    def complete(subargs_idx, subargs):
        """List valid subcommands for a command group that start with the last subarg."""
        if subargs_idx >= 1:
            return None
        incomplete = subargs[-1]
        print '\n'.join(cmd_group.list_commands({'parameters': [incomplete]}))
    return complete


@click.group()
@click.option('--profile', '-p')
@click.pass_context
def verdi(ctx, profile):
    """
    Toplevel command for click-implemented verdi commands.

    Might eventually replace ``execute_from_cmdline``, however, there is no way to directly call this command from the commandline
    currently. Instead, it is used for subcommand routing of commands written in click, see aiida/cmdline/commands/work.py for an
    example. In short it exists, so the name by which the subcommand is called ('verdi something something') matches it's command
    group hierarchy (group ``verdi``, subgroup ``something``, command ``something``).

    """
    ctx.obj = {'profile': profile}


@verdi.command()
@click.argument('completion_args', nargs=-1, type=click.UNPROCESSED)
def completion(completion_args):
    """
    Completion command alias for click-implemented verdi commands.

    Due to the roundabout process by which click-implemented verdi commands are called, pressing <Tab><Tab> on one of them tries to
    call aiida.cmdline.commands.verdi with subcommand completion. Therefore in order to enable the same behaviour as on older commands,
    such a subcommand with the same signature must exist and must run the same code with the same arguments.
    """
    from aiida.cmdline.verdilib import Completion
    Completion().run(*completion_args)


@verdi.group()
def export():
    pass


@verdi.group()
def work():
    pass


@verdi.group()
def user():
    pass

@verdi.group()
def node():
    pass

@verdi.group('data', entry_point_group='aiida.cmdline.data', cls=Pluginable)
def data_cmd():
    """Verdi data interface for plugin commands."""
    pass


@verdi.group('daemon')
def daemon_cmd():
    pass


@verdi.group('code')
def code_cmd():
    pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Generic tests that need the use of the DB
"""

from aiida.backends.testbase import AiidaTestCase
from aiida.common.exceptions import ModificationNotAllowed
from aiida.orm.node import Node



class TestCode(AiidaTestCase):
    """
    Test the Code class.
    """

    def test_code_local(self):
        import tempfile

        from aiida.orm.code import Code
        from aiida.common.exceptions import ValidationError

        code = Code(local_executable='test.sh')
        with self.assertRaises(ValidationError):
            # No file with name test.sh
            code.store()

        with tempfile.NamedTemporaryFile() as f:
            f.write("#/bin/bash\n\necho test run\n")
            f.flush()
            code.add_path(f.name, 'test.sh')

        code.store()
        self.assertTrue(code.can_run_on(self.computer))
        self.assertTrue(code.get_local_executable(), 'test.sh')
        self.assertTrue(code.get_execname(), 'stest.sh')

    def test_remote(self):
        import tempfile

        from aiida.orm.code import Code
        from aiida.orm.computer import Computer
        from aiida.common.exceptions import ValidationError

        with self.assertRaises(ValueError):
            # remote_computer_exec has length 2 but is not a list or tuple
            _ = Code(remote_computer_exec='ab')

        # invalid code path
        with self.assertRaises(ValueError):
            _ = Code(remote_computer_exec=(self.computer, ''))

        # Relative path is invalid for remote code
        with self.assertRaises(ValueError):
            _ = Code(remote_computer_exec=(self.computer, 'subdir/run.exe'))

        # first argument should be a computer, not a string
        with self.assertRaises(TypeError):
            _ = Code(remote_computer_exec=('localhost', '/bin/ls'))

        code = Code(remote_computer_exec=(self.computer, '/bin/ls'))
        with tempfile.NamedTemporaryFile() as f:
            f.write("#/bin/bash\n\necho test run\n")
            f.flush()
            code.add_path(f.name, 'test.sh')

        with self.assertRaises(ValidationError):
            # There are files inside
            code.store()

        # If there are no files, I can store
        code.remove_path('test.sh')
        code.store()

        self.assertEquals(code.get_remote_computer().pk, self.computer.pk)
        self.assertEquals(code.get_remote_exec_path(), '/bin/ls')
        self.assertEquals(code.get_execname(), '/bin/ls')

        self.assertTrue(code.can_run_on(self.computer.dbcomputer))
        self.assertTrue(code.can_run_on(self.computer))
        othercomputer = Computer(name='another_localhost',
                                 hostname='localhost',
                                 transport_type='local',
                                 scheduler_type='pbspro',
                                 workdir='/tmp/aiida').store()
        self.assertFalse(code.can_run_on(othercomputer))


class TestWfBasic(AiidaTestCase):
    """
    Tests for the workflows
    """

    def test_versioning_lowlevel(self):
        """
        Checks the versioning.
        """
        from aiida.workflows.test import WFTestEmpty

        w = WFTestEmpty().store()

        # Even if I stored many attributes, this should stay at 1
        self.assertEquals(w._dbworkflowinstance.nodeversion, 1)
        self.assertEquals(w.dbworkflowinstance.nodeversion, 1)
        self.assertEquals(w._dbworkflowinstance.nodeversion, 1)

        w.label = "label1"
        w.label = "label2"
        self.assertEquals(w._dbworkflowinstance.nodeversion, 3)
        self.assertEquals(w.dbworkflowinstance.nodeversion, 3)
        self.assertEquals(w._dbworkflowinstance.nodeversion, 3)

        w.description = "desc1"
        w.description = "desc2"
        w.description = "desc3"
        self.assertEquals(w._dbworkflowinstance.nodeversion, 6)
        self.assertEquals(w.dbworkflowinstance.nodeversion, 6)
        self.assertEquals(w._dbworkflowinstance.nodeversion, 6)


class TestGroups(AiidaTestCase):
    """
    Test groups.
    """

    def test_creation(self):
        from aiida.orm.group import Group

        n = Node()
        stored_n = Node().store()

        with self.assertRaises(ValueError):
            # No name specified
            g = Group()

        g = Group(name='testgroup')

        with self.assertRaises(ValueError):
            # Too many parameters
            g = Group(name='testgroup', not_existing_kwarg=True)

        with self.assertRaises(ModificationNotAllowed):
            # g unstored
            g.add_nodes(n)

        with self.assertRaises(ModificationNotAllowed):
            # g unstored
            g.add_nodes(stored_n)

        g.store()

        with self.assertRaises(ValueError):
            # n unstored
            g.add_nodes(n)

        g.add_nodes(stored_n)

        nodes = list(g.nodes)
        self.assertEquals(len(nodes), 1)
        self.assertEquals(nodes[0].pk, stored_n.pk)

        # To avoid to find it in further tests
        g.delete()

    def test_description(self):
        """
        Test the update of the description both for stored and unstored
        groups.
        """
        from aiida.orm.group import Group

        n = Node().store()

        g1 = Group(name='testgroupdescription1', description="g1").store()
        g1.add_nodes(n)

        g2 = Group(name='testgroupdescription2', description="g2")

        # Preliminary checks
        self.assertTrue(g1.is_stored)
        self.assertFalse(g2.is_stored)
        self.assertEquals(g1.description, "g1")
        self.assertEquals(g2.description, "g2")

        # Change
        g1.description = "new1"
        g2.description = "new2"

        # Test that the groups remained in their proper stored state and that
        # the description was updated
        self.assertTrue(g1.is_stored)
        self.assertFalse(g2.is_stored)
        self.assertEquals(g1.description, "new1")
        self.assertEquals(g2.description, "new2")

        # Store g2 and check that the description is OK
        g2.store()
        self.assertTrue(g2.is_stored)
        self.assertEquals(g2.description, "new2")

        # clean-up
        g1.delete()
        g2.delete()

    def test_add_nodes(self):
        """
        Test different ways of adding nodes
        """
        from aiida.orm.group import Group

        n1 = Node().store()
        n2 = Node().store()
        n3 = Node().store()
        n4 = Node().store()
        n5 = Node().store()
        n6 = Node().store()
        n7 = Node().store()
        n8 = Node().store()

        g = Group(name='test_adding_nodes')
        g.store()
        # Single node
        g.add_nodes(n1)
        # List of nodes
        g.add_nodes([n2, n3])
        # Single DbNode
        g.add_nodes(n4.dbnode)
        # List of DbNodes
        g.add_nodes([n5.dbnode, n6.dbnode])
        # List of Nodes and DbNodes
        g.add_nodes([n7, n8.dbnode])

        # Check
        self.assertEquals(set([_.pk for _ in [n1, n2, n3, n4, n5, n6, n7, n8]]),
                          set([_.pk for _ in g.nodes]))

        # Try to add a node that is already present: there should be no problem
        g.add_nodes(n1)
        self.assertEquals(set([_.pk for _ in [n1, n2, n3, n4, n5, n6, n7, n8]]),
                          set([_.pk for _ in g.nodes]))

        # Cleanup
        g.delete()

    def test_remove_nodes(self):
        """
        Test node removal
        """
        from aiida.orm.group import Group

        n1 = Node().store()
        n2 = Node().store()
        n3 = Node().store()
        n4 = Node().store()
        n5 = Node().store()
        n6 = Node().store()
        n7 = Node().store()
        n8 = Node().store()
        n_out = Node().store()

        g = Group(name='test_remove_nodes').store()

        # Add initial nodes
        g.add_nodes([n1, n2, n3, n4, n5, n6, n7, n8])
        # Check
        self.assertEquals(set([_.pk for _ in [n1, n2, n3, n4, n5, n6, n7, n8]]),
                          set([_.pk for _ in g.nodes]))

        # Remove a node that is not in the group: nothing should happen
        # (same behavior of Django)
        g.remove_nodes(n_out)
        # Re-check
        self.assertEquals(set([_.pk for _ in [n1, n2, n3, n4, n5, n6, n7, n8]]),
                          set([_.pk for _ in g.nodes]))

        # Remove one Node and check
        g.remove_nodes(n4)
        self.assertEquals(set([_.pk for _ in [n1, n2, n3, n5, n6, n7, n8]]),
                          set([_.pk for _ in g.nodes]))
        # Remove one DbNode and check
        g.remove_nodes(n7.dbnode)
        self.assertEquals(set([_.pk for _ in [n1, n2, n3, n5, n6, n8]]),
                          set([_.pk for _ in g.nodes]))
        # Remove a list of Nodes and check
        g.remove_nodes([n1, n8])
        self.assertEquals(set([_.pk for _ in [n2, n3, n5, n6]]),
                          set([_.pk for _ in g.nodes]))
        # Remove a list of Nodes and check
        g.remove_nodes([n1, n8])
        self.assertEquals(set([_.pk for _ in [n2, n3, n5, n6]]),
                          set([_.pk for _ in g.nodes]))
        # Remove a list of DbNodes and check
        g.remove_nodes([n2.dbnode, n5.dbnode])
        self.assertEquals(set([_.pk for _ in [n3, n6]]),
                          set([_.pk for _ in g.nodes]))

        # Remove a mixed list of Nodes and DbNodes and check
        g.remove_nodes([n3, n6.dbnode])
        self.assertEquals(set(),
                          set([_.pk for _ in g.nodes]))

        # Cleanup
        g.delete()

    def test_creation_from_dbgroup(self):
        from aiida.orm.group import Group

        n = Node().store()

        g = Group(name='testgroup_from_dbgroup')
        g.store()
        g.add_nodes(n)

        dbgroup = g.dbgroup

        with self.assertRaises(ValueError):
            # Cannot pass more parameters, even if valid, if
            # dbgroup is specified
            Group(dbgroup=dbgroup, name="test")

        gcopy = Group(dbgroup=dbgroup)

        self.assertEquals(g.pk, gcopy.pk)
        self.assertEquals(g.uuid, gcopy.uuid)

        # To avoid to find it in further tests
        g.delete()

    def test_name_desc(self):
        from aiida.orm.group import Group

        g = Group(name='testgroup2', description='some desc')
        self.assertEquals(g.name, 'testgroup2')
        self.assertEquals(g.description, 'some desc')
        self.assertTrue(g.is_user_defined)
        g.store()
        # Same checks after storing
        self.assertEquals(g.name, 'testgroup2')
        self.assertTrue(g.is_user_defined)
        self.assertEquals(g.description, 'some desc')

        # To avoid to find it in further tests
        g.delete()

    def test_delete(self):
        from aiida.orm.group import Group
        from aiida.common.exceptions import NotExistent

        n = Node().store()

        g = Group(name='testgroup3', description='some other desc')
        g.store()

        gcopy = Group.get(name='testgroup3')
        self.assertEquals(g.uuid, gcopy.uuid)

        g.add_nodes(n)
        self.assertEquals(len(g.nodes), 1)

        g.delete()

        with self.assertRaises(NotExistent):
            # The group does not exist anymore
            Group.get(name='testgroup3')

        # I should be able to restore it
        g.store()

        # Now, however, by deleting and recreating it, I lost the elements
        self.assertEquals(len(g.nodes), 0)
        self.assertEquals(g.name, 'testgroup3')
        self.assertEquals(g.description, 'some other desc')
        self.assertTrue(g.is_user_defined)

        # To avoid to find it in further tests
        g.delete()

    def test_rename(self):
        """
        Test the renaming of a Group
        """
        from aiida.orm.group import Group

        name_original = 'groupie'
        name_changed = 'nogroupie'

        g = Group(name=name_original, description='I will be renamed')

        # Check name changes work before storing
        self.assertEquals(g.name, name_original)
        g.name = name_changed
        self.assertEquals(g.name, name_changed)

        # Revert the name to its original and store it
        g.name = name_original
        g.store()

        # Check name changes work after storing
        self.assertEquals(g.name, name_original)
        g.name = name_changed
        self.assertEquals(g.name, name_changed)


class TestDbExtras(AiidaTestCase):
    """
    Test Extras
    """
    pass

    def test_replacement(self):
        pass

class TestBool(AiidaTestCase):
    def test_bool_conversion(self):
        from aiida.orm.data.bool import Bool
        for val in [True, False]:
            self.assertEqual(val, bool(Bool(val)))

    def test_int_conversion(self):
        from aiida.orm.data.bool import Bool
        for val in [True, False]:
            self.assertEqual(int(val), int(Bool(val)))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import unicode_literals

from django.db import models, migrations
import aiida.utils.timezone
from aiida.backends.djsite.db.migrations import update_schema_version



SCHEMA_VERSION = "1.0.3"


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0002_db_state_change'),
    ]

    operations = [
        migrations.AddField(
            model_name='dblink',
            name='type',
            field=models.CharField(db_index=True, max_length=255, blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbcalcstate',
            name='time',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbcomment',
            name='ctime',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbgroup',
            name='time',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dblock',
            name='creation',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dblog',
            name='time',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbnode',
            name='ctime',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbuser',
            name='date_joined',
            field=models.DateTimeField(default=aiida.utils.timezone.now),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbworkflow',
            name='ctime',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbworkflowdata',
            name='time',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='dbworkflowstep',
            name='time',
            field=models.DateTimeField(default=aiida.utils.timezone.now, editable=False),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='dblink',
            unique_together=set([]),
        ),
        update_schema_version(SCHEMA_VERSION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Add a column

Revision ID: b947a8821295
Revises: 470e57bc0936
Create Date: 2017-06-16 18:09:03.152982

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b947a8821295'
down_revision = '470e57bc0936'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('account', sa.Column('last_transaction_date', sa.DateTime))

def downgrade():
    op.drop_column('account', 'last_transaction_date')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.common.links import LinkType
from aiida.backends.testbase import AiidaTestCase
from aiida.orm.data.bool import get_true_node
from aiida.orm.data.int import Int
from aiida.orm.data.str import Str
from aiida.orm import load_node
from aiida.orm.calculation.function import FunctionCalculation
from aiida.work import run, run_get_node, submit, workfunction, Process, ExitCode

DEFAULT_INT = 256
DEFAULT_LABEL = 'Default label'
DEFAULT_DESCRIPTION = 'Default description'
CUSTOM_LABEL = 'Custom label'
CUSTOM_DESCRIPTION = 'Custom description'


class TestWf(AiidaTestCase):

    def setUp(self):
        super(TestWf, self).setUp()
        self.assertIsNone(Process.current())

        @workfunction
        def wf_return_input(inp):
            return {'result': inp}

        @workfunction
        def wf_return_true():
            return get_true_node()

        @workfunction
        def wf_args(a):
            return a

        @workfunction
        def wf_args_with_default(a=Int(DEFAULT_INT)):
            return a

        @workfunction
        def wf_kwargs(**kwargs):
            return kwargs

        @workfunction
        def wf_args_and_kwargs(a, **kwargs):
            result = {'a': a}
            result.update(kwargs)
            return result

        @workfunction
        def wf_args_and_default(a, b=Int(DEFAULT_INT)):
            return {'a': a, 'b': b}

        @workfunction
        def wf_default_label_description(a=Int(DEFAULT_INT), label=DEFAULT_LABEL, description=DEFAULT_DESCRIPTION):
            return a

        @workfunction
        def wf_exit_code(exit_status, exit_message):
            return ExitCode(exit_status.value, exit_message.value)

        @workfunction
        def wf_excepts(exception):
            raise RuntimeError(exception.value)

        self.wf_return_input = wf_return_input
        self.wf_return_true = wf_return_true
        self.wf_args = wf_args
        self.wf_args_with_default = wf_args_with_default
        self.wf_kwargs = wf_kwargs
        self.wf_args_and_kwargs = wf_args_and_kwargs
        self.wf_args_and_default = wf_args_and_default
        self.wf_default_label_description = wf_default_label_description
        self.wf_exit_code = wf_exit_code
        self.wf_excepts = wf_excepts

    def tearDown(self):
        super(TestWf, self).tearDown()
        self.assertIsNone(Process.current())

    def test_wf_varargs(self):
        """
        Variadic arguments are not supported and should raise
        """
        with self.assertRaises(ValueError):
            @workfunction
            def wf_varargs(*args):
                return get_true_node()

    def test_wf_args(self):
        """
        Simple workfunction that defines a single positional argument
        """
        INPUT = 1

        with self.assertRaises(ValueError):
            result = self.wf_args()

        result = self.wf_args(a=Int(INPUT))
        self.assertTrue(isinstance(result, Int))
        self.assertEquals(result, INPUT)

    def test_wf_args_with_default(self):
        """
        Simple workfunction that defines a single argument with a default
        """
        INPUT = 1

        result = self.wf_args_with_default()
        self.assertTrue(isinstance(result, Int))
        self.assertEquals(result, Int(DEFAULT_INT))

        result = self.wf_args_with_default(a=Int(INPUT))
        self.assertTrue(isinstance(result, Int))
        self.assertEquals(result, INPUT)

    def test_wf_kwargs(self):
        """
        Simple workfunction that defines keyword arguments
        """
        INPUT = {'a': Int(DEFAULT_INT)}

        result = self.wf_kwargs()
        self.assertTrue(isinstance(result, dict))
        self.assertEquals(result, {})

        result = self.wf_kwargs(**INPUT)
        self.assertTrue(isinstance(result, dict))
        self.assertEquals(result, INPUT)

    def test_wf_args_and_kwargs(self):
        """
        Simple workfunction that defines a positional argument and keyword arguments
        """
        ARGS_INPUT = (Int(DEFAULT_INT),)
        KWARGS_INPUT = {'b': Int(INPUT)}

        result = self.wf_args_and_kwargs(*ARGS_INPUT)
        self.assertTrue(isinstance(result, dict))
        self.assertEquals(result, {'a': ARGS_INPUT[0]})

        result = self.wf_args_and_kwargs(*ARGS_INPUT, **KWARGS_INPUT)
        self.assertTrue(isinstance(result, dict))
        self.assertEquals(result, {'a': ARGS_INPUT[0], 'b': Int(DEFAULT_INT)})

    def test_wf_args_and_kwargs(self):
        """
        Simple workfunction that defines a positional argument and an argument with a default
        """
        INPUT = 1
        ARGS_INPUT_DEFAULT = (Int(DEFAULT_INT),)
        ARGS_INPUT_EXPLICIT = (Int(DEFAULT_INT), Int(INPUT))

        result = self.wf_args_and_default(*ARGS_INPUT_DEFAULT)
        self.assertTrue(isinstance(result, dict))
        self.assertEquals(result, {'a': ARGS_INPUT_DEFAULT[0], 'b': Int(DEFAULT_INT)})

        result = self.wf_args_and_default(*ARGS_INPUT_EXPLICIT)
        self.assertTrue(isinstance(result, dict))
        self.assertEquals(result, {'a': ARGS_INPUT_EXPLICIT[0], 'b': ARGS_INPUT_EXPLICIT[1]})

    def test_wf_args_passing_kwargs(self):
        """
        A function that does not explicitly define kwargs, should not be able to be called with
        keyword arguments that were not explicitly defined
        """
        INPUT = 1

        with self.assertRaises(ValueError):
            result = self.wf_args(a=Int(INPUT), b=Int(INPUT))

    def test_wf_set_label_description(self):
        """
        Verify that the label and description can be set for all workfunction variants
        """
        result, node = self.wf_args.run_get_node(a=Int(DEFAULT_INT), label=CUSTOM_LABEL, description=CUSTOM_DESCRIPTION)
        self.assertEquals(node.label, CUSTOM_LABEL)
        self.assertEquals(node.description, CUSTOM_DESCRIPTION)

        result, node = self.wf_args_with_default.run_get_node(label=CUSTOM_LABEL, description=CUSTOM_DESCRIPTION)
        self.assertEquals(node.label, CUSTOM_LABEL)
        self.assertEquals(node.description, CUSTOM_DESCRIPTION)

        result, node = self.wf_kwargs.run_get_node(label=CUSTOM_LABEL, description=CUSTOM_DESCRIPTION)
        self.assertEquals(node.label, CUSTOM_LABEL)
        self.assertEquals(node.description, CUSTOM_DESCRIPTION)

        result, node = self.wf_args_and_kwargs.run_get_node(a=Int(DEFAULT_INT), label=CUSTOM_LABEL,
                                                            description=CUSTOM_DESCRIPTION)
        self.assertEquals(node.label, CUSTOM_LABEL)
        self.assertEquals(node.description, CUSTOM_DESCRIPTION)

        result, node = self.wf_args_and_default.run_get_node(a=Int(DEFAULT_INT), label=CUSTOM_LABEL,
                                                             description=CUSTOM_DESCRIPTION)
        self.assertEquals(node.label, CUSTOM_LABEL)
        self.assertEquals(node.description, CUSTOM_DESCRIPTION)

    def test_wf_default_label_description(self):
        """
        Verify that a workfunction can define a default label and description but they can be overriden by the
        user by explicitly passing them as keyword arguments
        """
        result, node = self.wf_default_label_description.run_get_node(a=Int(DEFAULT_INT))
        self.assertEquals(node.label, DEFAULT_LABEL)
        self.assertEquals(node.description, DEFAULT_DESCRIPTION)

        result, node = self.wf_default_label_description.run_get_node(label=CUSTOM_LABEL,
                                                                      description=CUSTOM_DESCRIPTION)
        self.assertEquals(node.label, CUSTOM_LABEL)
        self.assertEquals(node.description, CUSTOM_DESCRIPTION)

    def test_exit_status(self):
        """
        If a workfunction reaches the FINISHED process state, it has to have been successful
        which means that the exit status always has to be 0
        """
        result, node = self.wf_args_with_default.run_get_node()
        self.assertEquals(node.exit_status, 0)
        self.assertEquals(node.is_finished_ok, True)
        self.assertEquals(node.is_failed, False)

    def test_launchers(self):
        """
        Verify that the various launchers are working
        """
        result = run(self.wf_return_true)
        self.assertTrue(result)

        result, node = run_get_node(self.wf_return_true)
        self.assertTrue(result)
        self.assertEqual(result, get_true_node())
        self.assertTrue(isinstance(node, FunctionCalculation))

        with self.assertRaises(AssertionError):
            submit(self.wf_return_true)

    def test_return_exit_code(self):
        """
        A workfunction that returns an ExitCode namedtuple should have its exit status and message set FINISHED
        """
        exit_status = 418
        exit_message = 'I am a teapot'
        result, node = self.wf_exit_code.run_get_node(exit_status=Int(exit_status), exit_message=Str(exit_message))
        self.assertTrue(node.is_finished)
        self.assertFalse(node.is_finished_ok)
        self.assertEquals(node.exit_status, exit_status)
        self.assertEquals(node.exit_message, exit_message)

    def test_normal_exception(self):
        """
        If a process, for example a FunctionProcess, excepts, the exception should be stored in the node
        """
        exception = 'This workfunction excepted'

        with self.assertRaises(RuntimeError):
            result, node = self.wf_excepts.run_get_node(exception=Str(exception))
            self.assertTrue(node.is_excepted)
            self.assertEquals(node.exception, exception)

    def test_default_linkname(self):
        """
        Verify that a workfunction that returns a single Data node, a default link label
        will be used that can be accessed both through the getattr as through the getitem
        method of the OutputManager of the node
        """
        INPUT = 1

        result, node = run_get_node(self.wf_args, a=Int(INPUT))
        self.assertEquals(node.out.result, INPUT)
        self.assertEquals(getattr(node.out, Process.SINGLE_RETURN_LINKNAME), INPUT)
        self.assertEquals(node.out[Process.SINGLE_RETURN_LINKNAME], INPUT)

    def test_simple_workflow(self):
        @workfunction
        def add(a, b):
            return a + b

        @workfunction
        def mul(a, b):
            return a * b

        @workfunction
        def add_mul_wf(a, b, c):
            return mul(add(a, b), c)

        result = add_mul_wf(Int(3), Int(4), Int(5))

    def test_wf_links(self):
        @workfunction
        def a():
            pass

        @workfunction
        def b():
            return Int(5)

        @workfunction
        def c():
            a()
            return b()

        result, calc = run.get_node(c)

        self.assertIn(result.get_inputs(link_type=LinkType.CREATE)[0].pk, [c.pk for c in calc.called])
        self.assertEqual(calc.get_outputs(link_type=LinkType.RETURN)[0].pk, result.pk)

    def test_hashes(self):
        result, w1 = self.wf_return_input.run_get_node(inp=Int(2))
        result, w2 = self.wf_return_input.run_get_node(inp=Int(2))
        self.assertEqual(w1.get_hash(), w2.get_hash())

    def test_hashes_different(self):
        result, w1 = self.wf_return_input.run_get_node(inp=Int(2))
        result, w2 = self.wf_return_input.run_get_node(inp=Int(3))
        self.assertNotEqual(w1.get_hash(), w2.get_hash())

    def _check_hash_consistent(self, pid):
        wc = load_node(pid)
        self.assertEqual(wc.get_hash(), wc.get_extra('_aiida_hash'))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm import Data
from aiida.orm.data import to_aiida_type


class ParameterData(Data):
    """
    Pass as input in the init a dictionary, and it will get stored as internal
    attributes.

    Usual rules for attribute names apply (in particular, keys cannot start with
    an underscore). If this is the case, a ValueError will be raised.

    You can then change/delete/add more attributes before storing with the
    usual methods of aiida.orm.Node
    """

    def set_dict(self, dict):
        """
        Replace the current dictionary with another one.

        :param dict: The dictionary to set.
        """
        import copy
        from aiida.common.exceptions import ModificationNotAllowed

        old_dict = copy.deepcopy(self.get_dict())

        try:
            # Delete existing attributes
            self._del_all_attrs()
            # I set the keys
            self.update_dict(dict)
        except ModificationNotAllowed:
            # I reraise here to avoid to go in the generic 'except' below,
            # that would raise again the same exception
            raise
        except:
            # Try to restore the old data
            self._del_all_attrs()
            self.update_dict(old_dict)
            raise

    def update_dict(self, dict):
        """
        Update the current dictionary with the keys provided in the dictionary.

        :param dict: a dictionary with the keys to substitute. It works like
          dict.update(), adding new keys and overwriting existing keys.
        """
        for k, v in dict.iteritems():
            self._set_attr(k, v)

    def get_dict(self):
        """
        Return a dict with the parameters
        """
        return dict(self.iterattrs())

    def keys(self):
        """
        Iterator of valid keys stored in the ParameterData object
        """
        for k in self.attrs():
            yield k

    def add_path(self, *args, **kwargs):
        from aiida.common.exceptions import ModificationNotAllowed

        raise ModificationNotAllowed(
            "Cannot add files or directories to a ParameterData object")

        # def validate(self):
        #        # There should be nothing specific to check
        #        super(ParameterData,self).validate()

    @property
    def dict(self):
        """
        To be used to get direct access to the underlying dictionary with the
        syntax node.dict.key or node.dict['key'].

        :return: an instance of the AttributeResultManager.
        """
        from aiida.orm.node import AttributeManager

        return AttributeManager(self)


@to_aiida_type.register(dict)
def _(value):
    return ParameterData(dict=value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import os, tempfile

def draw_graph(origin_node, ancestor_depth=None, descendant_depth=None, format='dot',
        include_calculation_inputs=False, include_calculation_outputs=False):
    """
    The algorithm starts from the original node and goes both input-ward and output-ward via a breadth-first algorithm.

    :param origin_node: An Aiida node, the starting point for drawing the graph
    :param int ancestor_depth: The maximum depth of the ancestors drawn. If left to None, we recurse until the graph is fully explored
    :param int descendant_depth: The maximum depth of the descendants drawn. If left to None, we recurse until the graph is fully explored
    :param str format: The format, by default dot

    :returns: The exit_status of the os.system call that produced the valid file
    :returns: The file name of the final output

    ..note::
        If an invalid format is provided graphviz prints a helpful message, so this doesn't need to be implemented here.
    """
    # 
    # until the connected part of the graph that contains the root_pk is fully explored.
    # TODO this command deserves to be improved, with options and further subcommands

    from aiida.orm.calculation import Calculation
    from aiida.orm.calculation.job import JobCalculation
    from aiida.orm.code import Code
    from aiida.orm.node import Node
    from aiida.common.links import LinkType
    from aiida.orm.querybuilder import QueryBuilder

    def draw_node_settings(node, **kwargs):
        """
        Returns a string with all infos needed in a .dot file  to define a node of a graph.
        :param node:
        :param kwargs: Additional key-value pairs to be added to the returned string
        :return: a string
        """
        if isinstance(node, Calculation):
            shape = "shape=polygon,sides=4"
        elif isinstance(node, Code):
            shape = "shape=diamond"
        else:
            shape = "shape=ellipse"
        if kwargs:
            additional_params = ",{}".format(
                ",".join('{}="{}"'.format(k, v) for k, v in kwargs.iteritems()))
        else:
            additional_params = ""
        if node.label:
            label_string = "\n'{}'".format(node.label)
            additional_string = ""
        else:
            additional_string = "\n {}".format(node.get_desc())
            label_string = ""
        labelstring = 'label="{} ({}){}{}"'.format(
            node.__class__.__name__, node.pk, label_string,
            additional_string)
        return "N{} [{},{}{}];".format(node.pk, shape, labelstring,
                                       additional_params)

    def draw_link_settings(inp_id, out_id, link_label, link_type):
        if link_type in (LinkType.CREATE.value, LinkType.INPUT.value):
            style='solid'  # Solid lines and black colors
            color = "0.0 0.0 0.0" # for CREATE and INPUT (The provenance graph)
        elif link_type == LinkType.RETURN.value:
            style='dotted'  # Dotted  lines of
            color = "0.0 0.0 0.0" # black color for Returns
        elif link_type == LinkType.CALL.value:
            style='bold' # Bold lines and
            color = "0.0 1.0 1.0" # Bright red for calls
        else:
            style='solid'   # Solid and
            color="0.0 0.0 0.5" #grey lines for unspecified links!
        return '    {} -> {} [label="{}", color="{}", style="{}"];'.format("N{}".format(inp_id),  "N{}".format(out_id), link_label, color, style)

    # Breadth-first search of all ancestors and descendant nodes of a given node
    links = {}  # Accumulate links here
    nodes = {origin_node.pk: draw_node_settings(origin_node, style='filled', color='lightblue')} #Accumulate nodes specs here
    # Additional nodes (the ones added with either one of  include_calculation_inputs or include_calculation_outputs
    # is set to true. I have to put them in a different dictionary because nodes is the one used for the recursion,
    # whereas these should not be used for the recursion:
    additional_nodes = {}

    last_nodes = [origin_node] # Put the nodes whose links have not been scanned yet

    # Go through the graph on-ward (i.e. look at inputs)
    depth = 0
    while last_nodes:
        # I augment depth every time I get through a new iteration
        depth += 1
        # I check whether I should stop here:
        if ancestor_depth is not None and depth > ancestor_depth:
            break
        # I continue by adding new nodes here!
        new_nodes = []
        for node in last_nodes:
            # This query gives me all the inputs of this node, and link labels and types!
            input_query = QueryBuilder()
            input_query.append(Node, filters={'id':node.pk}, tag='n')
            input_query.append(Node, input_of='n', edge_project=('id', 'label', 'type'), project='*', tag='inp')
            for inp, link_id, link_label, link_type in input_query.iterall():
                # I removed this check, to me there is no way that this link was already referred to!
                # if link_id not in links:
                links[link_id] = draw_link_settings(inp.pk, node.pk, link_label, link_type)
                # For the nodes I need to check, maybe this same node is referred to multiple times.
                if inp.pk not in nodes:
                    nodes[inp.pk] = draw_node_settings(inp)
                    new_nodes.append(inp)

            # Checking whether I also should include all the outputs of a calculation into the drawing:
            if include_calculation_outputs and isinstance(node, Calculation):
                # Query for the outputs, giving me also link labels and types:
                output_query = QueryBuilder()
                output_query.append(Node, filters={'id':node.pk}, tag='n')
                output_query.append(Node, output_of='n', edge_project=('id', 'label', 'type'), project='*', tag='out')
                # Iterate through results
                for out, link_id, link_label, link_type in output_query.iterall():
                    # This link might have been drawn already, because the output is maybe
                    # already drawn.
                    # To check: Maybe it's more efficient not to check this, since
                    # the dictionaries are large and contain many keys...
                    # I.e. just always draw, also when overwriting an existing (identical) entry.
                    if link_id not in links:
                        links[link_id] = draw_link_settings(node.pk, out.pk, link_label, link_type)
                    if out.pk not in nodes and out.pk not in additional_nodes:
                        additional_nodes[out.pk] = draw_node_settings(out)

        last_nodes = new_nodes


    # Go through the graph down-ward (i.e. look at outputs)
    last_nodes = [origin_node]
    depth = 0
    while last_nodes:
        depth += 1
        # Also here, checking of maximum descendant depth is set and applies.
        if descendant_depth is not None and depth > descendant_depth:
            break
        new_nodes = []

        for node in last_nodes:
            # Query for the outputs:
            output_query = QueryBuilder()
            output_query.append(Node, filters={'id':node.pk}, tag='n')
            output_query.append(Node, output_of='n', edge_project=('id', 'label', 'type'), project='*', tag='out')

            for out, link_id, link_label, link_type in output_query.iterall():
                # Draw the link
                links[link_id] = draw_link_settings(node.pk, out.pk, link_label, link_type)
                if out.pk not in nodes:
                    nodes[out.pk] = draw_node_settings(out)
                    new_nodes.append(out)

            if include_calculation_inputs and isinstance(node, Calculation):
                input_query = QueryBuilder()
                input_query.append(Node, filters={'id':node.pk}, tag='n')
                input_query.append(Node, input_of='n', edge_project=('id', 'label', 'type'), project='*', tag='inp')
                for inp, link_id, link_label, link_type in input_query.iterall():
                    # Also here, maybe it's just better not to check?
                    if link_id not in links:
                        links[link_id] = draw_link_settings(inp.pk, node.pk, link_label, link_type)
                    if inp.pk not in nodes and inp.pk not in additional_nodes:
                        additional_nodes[inp.pk] = draw_node_settings(inp)
        last_nodes = new_nodes

    # Writing the graph to a temporary file
    fd, fname = tempfile.mkstemp(suffix='.dot')
    with open(fname, 'w') as fout:
        fout.write("digraph G {\n")
        for l_name, l_values in links.iteritems():
            fout.write('    {}\n'.format(l_values))
        for n_name, n_values in nodes.iteritems():
            fout.write("    {}\n".format(n_values))
        for n_name, n_values in additional_nodes.iteritems():
            fout.write("    {}\n".format(n_values))
        fout.write("}\n")

    # Now I am producing the output file
    output_file_name = "{0}.{format}".format(origin_node.pk, format=format)
    exit_status = os.system('dot -T{format} {0} -o {1}'.format(fname, output_file_name, format=format))
    # cleaning up by removing the temporary file
    os.remove(fname)
    return exit_status, output_file_name



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm import CalculationFactory
from aiida.parsers.parser import Parser
from aiida.orm.data.int import Int
from aiida.orm.data.parameter import ParameterData


ArithmeticAddCalculation = CalculationFactory('simpleplugins.arithmetic.add')


class ArithmeticAddParser(Parser):

    _linkname_output = 'sum'

    def __init__(self, calculation):
        """
        Initialize the Parser for an ArithmeticAddCalculation

        :param calculation: instance of the ArithmeticAddCalculation
        """
        if not isinstance(calculation, ArithmeticAddCalculation):
            raise ValueError('Input calculation must be of type {}'.format(type(ArithmeticAddCalculation)))

        self.calculation = calculation

        super(ArithmeticAddParser, self).__init__(calculation)

    @classmethod
    def get_linkname_output(self):
        """
        The name of the link used for the output node
        """
        return self._linkname_output

    def parse_with_retrieved(self, retrieved):
        """
        Parse the contents of the output file

        :param retrieved: a dictionary of retrieved nodes
        """
        is_success = True
        output_nodes = []

        try:
            output_folder = retrieved[self.calculation._get_linkname_retrieved()]
        except KeyError:
            self.logger.error("no retrieved folder found")
            return False, ()

        # Verify the standard output file is present, parse the value and attach as output node
        try:
            filepath_stdout = output_folder.get_abs_path(self.calculation._OUTPUT_FILE_NAME)
        except OSError as exception:
            self.logger.error("expected output file '{}' was not found".format(filepath_stdout))
            return False, ()

        is_success, output_node = self.parse_stdout(filepath_stdout)

        if not is_success:
            self.logger.error('failed to parse the result from the output file {}'.format(filepath_stdout))
            return False, ()

        output_nodes.append((self.get_linkname_output(), output_node))

        return 0, output_nodes

    def parse_stdout(self, filepath):
        """
        Parse the sum from the output of the ArithmeticAddcalculation written to standard out

        :param filepath: path to file containing output written to stdout
        :returns: boolean representing success status of parsing, True equals parsing was successful
        :returns: the sum as a node
        """
        is_successful = True

        try:
            with open(filepath, 'r') as handle:
                output = handle.read()
        except IOError:
            return False, None

        try:
            result = int(output)
        except ValueError:
            return False, None

        output_node = Int(result)

        return is_successful, output_node



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import os

from django.db import transaction

from aiida.orm.implementation.general.code import AbstractCode
from aiida.orm.implementation import Computer
from aiida.common.exceptions import (NotExistent, MultipleObjectsError,
                                     InvalidOperation)



class Code(AbstractCode):

    @classmethod
    def get(cls, pk=None, label=None, machinename=None):
        return super(Code, cls).get(pk, label, machinename)

    @classmethod
    def get_from_string(cls, code_string):
        return super(Code, cls).get_from_string(code_string)

    @classmethod
    def list_for_plugin(cls, plugin, labels=True):
        """
        Return a list of valid code strings for a given plugin.

        :param plugin: The string of the plugin.
        :param labels: if True, return a list of code names, otherwise
          return the code PKs (integers).
        :return: a list of string, with the code names if labels is True,
          otherwise a list of integers with the code PKs.
        """
        valid_codes = list(cls.query(
            dbattributes__key="input_plugin",
            dbattributes__tval=plugin))

        if labels:
            return [c.label for c in valid_codes]
        else:
            return [c.pk for c in valid_codes]

    def set_remote_computer_exec(self, remote_computer_exec):
        """
        Set the code as remote, and pass the computer on which it resides
        and the absolute path on that computer.

        Args:
            remote_computer_exec: a tuple (computer, remote_exec_path), where
              computer is a aiida.orm.Computer or an
              aiida.backends.djsite.db.models.DbComputer object, and
              remote_exec_path is the absolute path of the main executable on
              remote computer.
        """
        from aiida.backends.djsite.db.models import DbComputer
        if (not isinstance(remote_computer_exec, (list, tuple))
            or len(remote_computer_exec) != 2):
            raise ValueError("remote_computer_exec must be a list or tuple "
                             "of length 2, with machine and executable "
                             "name")

        computer, remote_exec_path = tuple(remote_computer_exec)

        if not os.path.isabs(remote_exec_path):
            raise ValueError(
                "exec_path must be an absolute path (on the remote machine)")

        remote_dbcomputer = computer
        if isinstance(remote_dbcomputer, Computer):
            remote_dbcomputer = remote_dbcomputer.dbcomputer
        if not (isinstance(remote_dbcomputer, DbComputer)):
            raise TypeError(
                "computer must be either a Computer or DbComputer object")

        self._set_remote()

        self.dbnode.dbcomputer = remote_dbcomputer
        self._set_attr('remote_exec_path', remote_exec_path)

    def _set_local(self):
        """
        Set the code as a 'local' code, meaning that all the files belonging to the code
        will be copied to the cluster, and the file set with set_exec_filename will be
        run.

        It also deletes the flags related to the local case (if any)
        """
        self._set_attr('is_local', True)
        self.dbnode.dbcomputer = None
        try:
            self._del_attr('remote_exec_path')
        except AttributeError:
            pass

    def can_run_on(self, computer):
        """
        Return True if this code can run on the given computer, False otherwise.

        Local codes can run on any machine; remote codes can run only on the machine
        on which they reside.

        TODO: add filters to mask the remote machines on which a local code can run.
        """
        from aiida.backends.djsite.db.models import DbComputer
        if self.is_local():
            return True
        else:
            dbcomputer = computer
            if isinstance(dbcomputer, Computer):
                dbcomputer = dbcomputer.dbcomputer
            if not isinstance(dbcomputer, DbComputer):
                raise ValueError(
                    "computer must be either a Computer or DbComputer object")
            dbcomputer = DbComputer.get_dbcomputer(computer)
            return (dbcomputer.pk ==
                    self.get_remote_computer().dbcomputer.pk)




def delete_code(code):
    """
    Delete a code from the DB.
    Check before that there are no output nodes.

    NOTE! Not thread safe... Do not use with many users accessing the DB
    at the same time.

    Implemented as a function on purpose, otherwise complicated logic would be
    needed to set the internal state of the object after calling
    computer.delete().
    """
    if not isinstance(code, Code):
        raise TypeError("code must be an instance of "
                        "aiida.orm.computer.Code")

    existing_outputs = code.get_outputs()

    if len(existing_outputs) != 0:
        raise InvalidOperation("Unable to delete the requested code because it "
                               "has {} output links".format(
            len(existing_outputs)))
    else:
        repo_folder = code._repository_folder
        with transaction.atomic():
            code.dbnode.delete()
            repo_folder.erase()
